# me2.plugs.wave/gadget.py
#
#

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.lib.persist import PlugPersist

gadgeturls = PlugPersist('gadgeturls')

gadgeturls.data['gadget'] = 'https://botme2.appspot.com/gadget.xml'
gadgeturls.data['poll'] = 'https://botme2.appspot.com/poll.xml'
gadgeturls.data['iframe'] = 'https://botme2.appspot.com/iframe.xml'
gadgeturls.data['loadiframe'] = 'https://botme2.appspot.com/loadiframe.xml'

def loadroot(event, url):
    if event.rootblip:
        from waveapi import element
        event.rootblip.append(element.Gadget(url))
        return True
    else:
        event.reply("can't find root blip.")
        return False

def load(event, url):
    if event.blip:
        from waveapi import element
        event.blip.append(element.Gadget(url))
        return True
    else:
        event.reply("can't find root blip.")
        return False

def handle_gadgetload(bot, event):
    if event.bottype != "wave":
        event.reply("this command only works in google wave.");
        return

    if not event.rest:
        event.missing('<gadgetname>')
        return

    try:
        url = gadgeturls.data[event.rest]
        if load(event, url):
            event.reply('loaded %s' % url)
    except KeyError:
        event.reply("we don't have a url for %s" % event.rest)

cmnds.add("gadget-load", handle_gadgetload, 'USER')
examples.add("gadget-load", "load a gadget into a blip", "gadget-load")

def handle_gadgetloadroot(bot, event):
    if event.bottype != "wave":
        event.reply("this command only works in google wave.");
        return

    if not event.rest:
        event.missing('<gadgetname>')
        return

    try:
        url = gadgeturls.data[event.rest]
        if loadroot(event, url):
            event.reply('loaded %s' % url)
    except KeyError:
        event.reply("we don't have a url for %s" % event.rest)

cmnds.add("gadget-loadroot", handle_gadgetloadroot, 'USER')
examples.add("gadget-loadroot", "load a gadget into the root blip", "gadget-loadroot")

def handle_gadgetiframe(bot, event):
    if event.bottype != "wave":
        event.reply("this command only works in google wave.");
        return

    if not event.rest:
        event.missing('<url>')
        return

    try:
        url = gadgeturls.data['loadiframe'] + "?&iframeurl=%s" % event.rest
        event.reply('loading %s' % url)
        load(event, url)
    except KeyError:
        event.reply("we don't have a iframe url")

cmnds.add("gadget-iframe", handle_gadgetiframe, 'USER')
examples.add("gadget-iframe", "load a url into a iframe", "gadget-iframe")

def handle_gadgetaddurl(bot, event):

    try:
        (name, url) = event.args
    except ValueError:
        event.missing('<name> <url>')
        return

    if not gadgeturls.data.has_key(name):
        gadgeturls.data[name] = url
        gadgeturls.save()
    else:
        event.reply("we already have a %s gadget" % name)

cmnds.add("gadget-addurl", handle_gadgetaddurl, 'USER')
examples.add("gadget-addurl", "store a gadget url", "gadget-addurl me2 https://botme2.appspot.com/iframe.xml")

def handle_gadgetdelurl(bot, event):

    try:
        (name, url) = event.args
    except ValueError:
        event.missing('<name> <url>')
        return

    gadgeturls.data[name] = url
    gadgeturls.save()

cmnds.add("gadget-delurl", handle_gadgetdelurl, 'OPER')
examples.add("gadget-delurl", "delete a gadget url", "gadget-delurl mygadget")

def handle_gadgetlist(bot, event):
    result = []

    for name, url in gadgeturls.data.iteritems():
        result.append("%s - %s" % (name, url))

    event.reply("available gadgets: ", result)

cmnds.add("gadget-list", handle_gadgetlist, 'USER')
examples.add("gadget-list", "list known gadget urls", "gadget-list")


def handle_gadgetconsole(bot, event):
    if event.bottype != "wave":
        event.reply("this command only works in google wave.");
        return

    wave = event.chan
    if wave.data.feeds and wave.data.dotitle:
        event.set_title("BOTME2 - %s #%s" % (" - ".join(wave.data.feeds), str(wave.data.nrcloned)))

    from waveapi import element

    #url = gadgeturls.data['loadiframe'] + "?&iframeurl=https://botme2.appspot.com"
    #event.reply('loading %s' % url)
    event.append("loading ...\n")
    #load(event, "http://botme2.appspot.com/iframe.xml")

    event.append(
        element.Gadget('http://botme2.appspot.com/console.xml?gadget_cache=0'))

cmnds.add("gadget-console", handle_gadgetconsole, 'OPER')
examples.add("gadget-console", "load the console gadget", "gadget-console")



# me2/botbase.py
#
#

""" base class for all bots. """

## me2 imports

from me2.utils.exception import handle_exception
from runner import defaultrunner
from eventhandler import mainhandler
from me2.utils.lazydict import LazyDict
from plugins import plugs as coreplugs
from callbacks import callbacks, first_callbacks, last_callbacks, remote_callbacks
from eventbase import EventBase
from errors import NoSuchCommand, PlugsNotConnected, NoOwnerSet, NameNotSet, NoEventProvided
from commands import Commands
from config import Config
from me2.utils.pdod import Pdod
from channelbase import ChannelBase
from less import Less, outcache
from boot import boot, getcmndperms, default_plugins
from me2.utils.locking import lockdec
from exit import globalshutdown
from me2.utils.generic import splittxt, toenc, fromenc, waitforqueue, strippedtxt
from me2.utils.trace import whichmodule
from fleet import getfleet
from me2.utils.name import stripname
from tick import tickloop
from threads import start_new_thread, threaded
from gatekeeper import GateKeeper

## basic imports

import time
import logging
import copy
import sys
import getpass
import os
import thread
import types
import threading
import Queue
import re

## defines

cpy = copy.deepcopy

## locks

reconnectlock = threading.RLock()
reconnectlocked = lockdec(reconnectlock)

lock = thread.allocate_lock()
locked = lockdec(lock)

## classes

class BotBase(LazyDict):

    """ base class for all bots. """

    def __init__(self, cfg=None, usersin=None, plugs=None, botname=None, nick=None, *args, **kwargs):
        if not botname and cfg and cfg.botname: botname = cfg.botname
        if botname: self.botname = botname
        else: self.botname = u"default-%s" % str(type(self)).split('.')[-1][:-2]
        logging.info("botbase - name is %s" % self.botname)
        self.fleetdir = u'fleet' + os.sep + stripname(self.botname)
        if cfg: self.cfg = Config(self.fleetdir + os.sep + u'config', input=cfg)
        else: self.cfg = Config(self.fleetdir + os.sep + u'config')
        LazyDict.__init__(self)
        self.update(self.cfg)
        self.curevent = None
        self.inqueue = Queue.Queue()
        self.outqueue = Queue.Queue()
        self.reconnectcount = 0
        self.stopped = False
        self.plugs = plugs or coreplugs
        self.gatekeeper = GateKeeper(self.botname)
        self.gatekeeper.allow(self.user or self.jid or self.server or self.botname)
        self.closed = False
        try:
            import waveapi
            self.isgae = True
            logging.debug("botbase - bot is a GAE bot (%s)" % self.botname)
        except ImportError:
            self.isgae = False
            logging.debug("botbase - bot is a shell bot (%s)" % self.botname)
        self.starttime = time.time()
        self.type = "base"
        self.status = "init"
        self.networkname = self.cfg.networkname or self.botname or ""
        if not self.uuid:
            if self.cfg and self.cfg.uuid: self.uuid = self.cfg.uuid
            else:
                self.uuid = self.cfg.uuid = uuid.uuid4()
                self.cfg.save()
        if self.cfg and not self.cfg.followlist: self.cfg.followlist = [] ; self.cfg.save()
        from me2.lib.datadir import getdatadir
        datadir = getdatadir()
        self.datadir = datadir + os.sep + self.fleetdir
        self.name = self.botname
        self.owner = self.cfg.owner
        if not self.owner:
            logging.debug(u"owner is not set in %s - using mainconfig" % self.cfg.cfile)
            self.owner = Config().owner
        self.setusers(usersin)
        logging.info(u"botbase - owner is %s" % self.owner)
        self.users.make_owner(self.owner)
        self.outcache = outcache
        self.userhosts = {}
        self.connectok = threading.Event()
        if not self.nick: self.nick = (nick or self.cfg.nick or u'me2')
        try:
            if not os.isdir(self.datadir): os.mkdir(self.datadir)
        except: pass
        self.setstate()
        self.stopreadloop = False
        self.stopoutloop = False
        self.outputlock = thread.allocate_lock()
        self.outqueues = [Queue.Queue() for i in range(10)]
        self.tickqueue = Queue.Queue()
        self.encoding = self.cfg.encoding or "utf-8"
        self.cmndperms = getcmndperms()
        fleet = getfleet(datadir)
        if not fleet.byname(self.name): fleet.bots.append(self) ; 
        if not self.isgae:
            defaultrunner.start()
            tickloop.start(self)

    def copyin(self, data):
        self.update(data)

    def _resumedata(self):
       """ overload this. """ 
       pass

    def enable(self, modname):
        """ enable plugin given its modulename. """
        try: self.cfg.blacklist and self.cfg.blacklist.remove(modname)
        except ValueError: pass           
        if self.cfg.loadlist and modname not in self.cfg.loadlist: self.cfg.loadlist.append(modname)
        self.cfg.save()

    def disable(self, modname):
        """ disable plugin given its modulename. """
        if self.cfg.blacklist and modname not in self.cfg.blacklist: self.cfg.blacklist.append(modname)
        if self.cfg.loadlist and modname in self.cfg.loadlist: self.cfg.loadlist.remove(modname)
        self.cfg.save()

    def put(self, event):
        """ put an event on the worker queue. """
        self.inqueue.put_nowait(event)

    def broadcast(self, txt):
        """ broadcast txt to all joined channels. """
        for chan in self.state['joinedchannels']:
            self.say(chan, txt)

    def _eventloop(self):
        """ fetch events from the inqueue and handle them. """
        logging.debug("%s - eventloop started" % self.name)
        while not self.stopped:
            event = self.inqueue.get()
            if not event: break
            self.doevent(event)
        logging.debug("%s - eventloop stopped" % self.name)

    def _getqueue(self):
        """ get one of the outqueues. """
        go = self.tickqueue.get()
        for index in range(len(self.outqueues)):
            if not self.outqueues[index].empty(): return self.outqueues[index]

    def _outloop(self):
        """ output loop. """
        logging.debug('%s - starting output loop' % self.name)
        self.stopoutloop = 0
        while not self.stopped and not self.stopoutloop:
            queue = self._getqueue()
            if queue:
                try:
                    res = queue.get() 
                except Queue.Empty: continue
                if not res: continue
                if not self.stopped and not self.stopoutloop:
                    logging.debug("%s - OUT - %s - %s" % (self.name, self.type, str(res))) 
                    self.out(*res)
            time.sleep(0.1)
        logging.debug('%s - stopping output loop' % self.name)

    def putonqueue(self, nr, *args):
        """ put output onto one of the output queues. """
        self.outqueues[10-nr].put_nowait(args)
        self.tickqueue.put_nowait('go')

    def outputsizes(self):
        """ return sizes of output queues. """
        result = []  
        for q in self.outqueues:
            result.append(q.qsize())
        return result

    def setstate(self, state=None):
        """ set state on the bot. """
        self.state = state or Pdod(self.datadir + os.sep + 'state')
        if self.state and not 'joinedchannels' in self.state.data: self.state.data.joinedchannels = []

    def setusers(self, users=None):
        """ set users on the bot. """
        if users:
            self.users = users
            return
        import me2.lib.users as u
        if not u.users: u.users_boot()
        self.users = u.users

    def loadplugs(self, packagelist=[]):
        """ load plugins from packagelist. """
        self.plugs.loadall(packagelist)
        return self.plugs

    def joinchannels(self):
        """ join channels. """
        time.sleep(Config().waitforjoin or 5)
        for i in self.state['joinedchannels']:
            try:
                logging.debug("%s - joining %s" % (self.name, i))
                channel = ChannelBase(i)
                if channel: key = channel.getpass()
                else: key=None
                start_new_thread(self.join, (i, key))
                time.sleep(1)
            except Exception, ex:
                logging.warn('%s - failed to join %s: %s' % (self.name, i, str(ex)))
                handle_exception()

    def start(self, connect=True):
        """ start the mainloop of the bot. """
        if not self.isgae: 
            if connect: self.connect()
            start_new_thread(self._outloop, ())
            start_new_thread(self._eventloop, ())
            start_new_thread(self._readloop, ())
            if connect:
                self.connectok.wait(120)
                if self.connectok.isSet():
                    logging.warn('%s - logged on !' % self.name)
                    start_new_thread(self.joinchannels, ())
                else: logging.warn("%s - failed to logon - connectok is not set" % self.name)
        self.status == "running"
        self.dostart(self.botname, self.type)

    def doremote(self, event):
        """ dispatch an event. """
        if not event: raise NoEventProvided()
        logging.info("======== start handling REMOTE event ========")
        event.prepare(self)
        self.status = "callback"
        starttime = time.time()
        msg = "%s - %s - %s - %s" % (self.name, event.auth, event.how, event.cbtype)
        if event.how == "background": logging.debug(msg)
        else: logging.info(msg)
        logging.debug("botbase - remote - %s" % event.dump())
        if self.closed:
            if self.gatekeeper.isblocked(event.origin): return
        if event.status == "done":
            logging.debug("%s - event is done .. ignoring" % self.name)
            return
        self.reloadcheck(event)
        e0 = cpy(event)
        e0.speed = 1
        remote_callbacks.check(self, e0)
        logging.info("======== start handling REMOTE event ========")
        return

    #@locked
    def doevent(self, event):
        """ dispatch an event. """
        if not event: raise NoEventProvided()
        if event.isremote(): self.doremote(event) ; return
        msg = "%s - %s - %s - %s" % (self.name, event.auth, event.how, event.cbtype)
        if event.cbtype in ['NOTICE']: logging.warn("%s - %s - %s" % (self.name, event.nick, event.txt))
        else:
            try:
                int(event.cbtype)
                logging.debug("======== start handling local event ========")
                logging.debug(msg)
            except ValueError:
                if event.cbtype in ['PING', 'PRESENCE'] or event.how == "background": 
                    logging.debug("======== start handling local event ========")
                    logging.debug(msg)
                else: logging.info("======== start handling local event ========") ; logging.info(msg)
        logging.debug(event.dump())
        event.prepare(self)
        self.status = "callback"
        starttime = time.time()
        if self.closed:
            if self.gatekeeper.isblocked(event.origin): return
        if event.status == "done":
            logging.debug("%s - event is done .. ignoring" % self.name)
            return
        self.reloadcheck(event)
        if event.msg or event.isdcc: event.speed = 2
        e1 = cpy(event)
        first_callbacks.check(self, e1)
        if not e1.stop: 
            callbacks.check(self, e1)
            if not e1.stop: last_callbacks.check(self, e1)
        event.callbackdone = True
        return event

    def ownercheck(self, userhost):
        """ check if provided userhost belongs to an owner. """
        if self.cfg and self.cfg.owner:
            if userhost in self.cfg.owner: return True
        return False

    def exit(self, stop=True):
        """ exit the bot. """ 
        logging.warn("%s - exit" % self.name)
        if not self.stopped: self.quit()
        if stop:
            self.stopped = True   
            self.stopreadloop = True  
            self.connected = False
        self.put(None)
        self.tickqueue.put_nowait('go')
        self.outqueue.put_nowait(None)
        self.shutdown()
        self.save()

    def _raw(self, txt, *args, **kwargs):
        """ override this. """ 
        logging.debug(u"%s - out - %s" % (self.name, txt))
        print txt

    def makeoutput(self, printto, txt, result=[], nr=375, extend=0, dot=", ", *args, **kwargs):
        if not txt: return ""
        txt = self.makeresponse(txt, result, dot)
        res1, nritems = self.less(printto, txt, nr+extend)
        return res1

    def out(self, printto, txt, how="msg", event=None, origin=None, html=False, *args, **kwargs):
        self.outnocb(printto, txt, how, event=event, origin=origin, html=html, *args, **kwargs)
        self.outmonitor(origin, printto, txt, event=event)

    write = out

    def outnocb(self, printto, txt, how="msg", event=None, origin=None, html=False, *args, **kwargs):
        self._raw(txt)

    writenocb = outnocb

    def say(self, channel, txt, result=[], how="msg", event=None, nr=375, extend=0, dot=", ", *args, **kwargs):
        txt = self.makeoutput(channel, txt, result, nr, extend, dot, *args, **kwargs)
        if txt: self.out(channel, txt, how, event=event, origin=channel, *args, **kwargs)
        if event: event.outqueue.put_nowait(txt)
        
    def saynocb(self, channel, txt, result=[], how="msg", event=None, nr=375, extend=0, dot=", ", *args, **kwargs):
        txt = self.makeoutput(channel, txt, result, nr, extend, dot, *args, **kwargs)
        if txt: self.outnocb(channel, txt, event=event, origin=channel, *args, **kwargs)

    def less(self, printto, what, nr=365):
        """ split up in parts of <nr> chars overflowing on word boundaries. """
        if type(what) == types.ListType: txtlist = what
        else:
            what = what.strip()
            txtlist = splittxt(what, nr)
        size = 0
        if not txtlist:   
            logging.debug("can't split txt from %s" % what)
            return ["", ""]
        res = txtlist[0]
        length = len(txtlist)
        if length > 1:
            logging.info("addding %s lines to %s outputcache" % (len(txtlist), printto))
            outcache.set(u"%s-%s" % (self.name, printto), txtlist[1:])
            res += "<b> - %s more</b>" % (length - 1) 
        return [res, length]

    def join(self, channel, password, *args, **kwargs):
        """ join a channel. """
        pass

    def part(self, channel, *args, **kwargs):
        """ leave a channel. """
        pass

    def action(self, channel, txt, *args, **kwargs):
        """ send action to channel. """
        pass

    def reconnect(self):
        """ reconnect to the server. """
        try:
            try: self.exit()
            except Exception, ex: handle_exception()
            self.reconnectcount += 1
            logging.warn('%s - reconnecting .. sleeping %s seconds' % (self.name, self.reconnectcount*15))
            time.sleep(self.reconnectcount * 15)   
            self.doreconnect()
        except Exception, ex: 
            handle_exception()

    def doreconnect(self):
        self.start(connect=True)

    def invite(self, *args, **kwargs):
        """ invite another user/bot. """
        pass

    def donick(self, nick, *args, **kwargs):
        """ do a nick change. """
        pass

    def shutdown(self, *args, **kwargs):
        """ shutdown the bot. """
        pass

    def quit(self, reason="", *args, **kwargs):
        """ close connection with the server. """
        pass

    def connect(self, reconnect=True, *args, **kwargs):
        """ connect to the server. """
        pass

    def names(self, channel, *args, **kwargs):
        """ request all names of a channel. """
        pass

    def save(self, *args, **kwargs):
        """ save bot state if available. """
        if self.state: self.state.save()

    def makeresponse(self, txt, result=[], dot=", ", *args, **kwargs):
        """ create a response from a string and result list. """
        res = []
        if result:
            txt = u"<b>" + txt + u"</b>"   
            for i in result:
                if not i: continue
                if type(i) == types.ListType or type(i) == types.TupleType:
                    try:
                        res.append(dot.join(i))
                    except TypeError: res.extend(i)
                else: res.append(i)
        if txt: return unicode(txt) + dot.join(res)   
        elif res: return dot.join(res)
        return ""
    
    def reloadcheck(self, event):
        """ check if plugin need to be reloaded for callback, """
        plugloaded = []
        target = event.cbtype or event.cmnd
        try:
            from boot import getcallbacktable   
            p = getcallbacktable()[target]
        except KeyError:
            logging.debug("botbase - can't find plugin to reload for %s" % event.cmnd)
            return
        logging.debug("%s - checking %s" % (self.name, unicode(p)))
        for name in p:
            if name in self.plugs: continue
            if name in default_plugins: pass
            elif self.cfg.blacklist and name in self.cfg.blacklist: continue
            elif self.cfg.loadlist and name not in self.cfg.loadlist: continue
            logging.warn("botbase - on demand reloading of %s" % name)
            try:
                mod = self.plugs.reload(name, force=True, showerror=False)
                if mod: plugloaded.append(mod) ; continue
            except Exception, ex: handle_exception(event)
        return plugloaded

    def send(self, *args, **kwargs):
        pass

    def sendnocb(self, *args, **kwargs):
        pass

    def normalize(self, what):
        """ convert markup to IRC bold. """
        txt = strippedtxt(what)
        txt = re.sub("\s+", " ", what)
        txt = txt.replace("<b>", "\002")
        txt = txt.replace("</b>", "\002")
        txt = txt.replace("<i>", "")
        txt = txt.replace("</i>", "")
        txt = txt.replace("&lt;b&gt;", "\002")
        txt = txt.replace("&lt;/b&gt;", "\002")
        txt = txt.replace("&lt;i&gt;", "")
        txt = txt.replace("&lt;/i&gt;", "")
        txt = txt.replace("&lt;h2&gt;", "\002")
        txt = txt.replace("&lt;/h2&gt;", "\002")
        txt = txt.replace("&lt;h3&gt;", "\002")
        txt = txt.replace("&lt;/h3&gt;", "\002")
        txt = txt.replace("&lt;li&gt;", "\002")
        txt = txt.replace("&lt;/li&gt;", "\002")
        return txt

    def dostart(self, botname, bottype, *args, **kwargs):
        """ create an START event and send it to callbacks. """
        e = EventBase()
        e.bot = self
        e.botname = botname
        e.bottype = bottype
        e.origin = botname
        e.ruserhost = self.botname +'@' + self.uuid
        e.userhost = e.ruserhost
        e.channel = botname
        e.origtxt = str(time.time())
        e.txt = e.origtxt
        e.cbtype = 'START'
        e.botoutput = False
        e.ttl = 1
        e.nick = self.nick or self.botname
        self.doevent(e)
        logging.debug("%s - START event send to callbacks" % botname)

    def outmonitor(self, origin, channel, txt, event=None):
        """ create an OUTPUT event with provided txt and send it to callbacks. """
        if event: e = cpy(event)
        else: e = EventBase()
        if e.status == "done":
            logging.debug("%s - outmonitor - event is done .. ignoring" % self.name)
            return
        e.bot = self
        e.origin = origin
        e.ruserhost = self.botname +'@' + self.uuid
        e.userhost = e.ruserhost
        e.channel = channel
        e.origtxt = txt
        e.txt = txt
        e.cbtype = 'OUTPUT'
        e.botoutput = True
        e.nodispatch = True
        e.ttl = 1
        e.nick = self.nick or self.botname
        e.bind(self)
        first_callbacks.check(self, e)

    def docmnd(self, origin, channel, txt, event=None, wait=0):
        """ do a command. """
        if event: e = cpy(event)
        else: e = EventBase()
        e.bot = self
        e.origin = origin
        e.ruserhost = origin
        e.auth = origin
        e.userhost = origin
        e.channel = channel
        e.txt = unicode(txt)
        e.nick = e.userhost.split('@')[0]
        e.usercmnd = e.txt.split()[0]
        e.iscmnd = True
        e.iscallback = False
        e.allowqueues = True
        e.onlyqueues = False
        e.closequeue = True
        if wait: e.direct = True
        e.bind(self)
        try:
            result = self.plugs.dispatch(self, e, wait=wait)
            return result
        except NoSuchCommand: e.reply("no such command: %s" % e.usercmnd)



#!/usr/bin/python2.4
#
# Copyright (C) 2009 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the element module."""


import base64
import unittest

import element
import util


class TestElement(unittest.TestCase):
  """Tests for the element.Element class."""

  def testProperties(self):
    el = element.Element(element.Gadget.class_type,
                         key='value')
    self.assertEquals('value', el.key)

  def testFormElement(self):
    el = element.Input('input')
    self.assertEquals(element.Input.class_type, el.type)
    self.assertEquals(el.value, '')
    self.assertEquals(el.name, 'input')

  def testImage(self):
    image = element.Image('http://test.com/image.png', width=100, height=100)
    self.assertEquals(element.Image.class_type, image.type)
    self.assertEquals(image.url, 'http://test.com/image.png')
    self.assertEquals(image.width, 100)
    self.assertEquals(image.height, 100)

  def testAttachment(self):
    attachment = element.Attachment(caption='My Favorite', data='SomefakeData')
    self.assertEquals(element.Attachment.class_type, attachment.type)
    self.assertEquals(attachment.caption, 'My Favorite')
    self.assertEquals(attachment.data, 'SomefakeData')

  def testGadget(self):
    gadget = element.Gadget('http://test.com/gadget.xml')
    self.assertEquals(element.Gadget.class_type, gadget.type)
    self.assertEquals(gadget.url, 'http://test.com/gadget.xml')

  def testInstaller(self):
    installer = element.Installer('http://test.com/installer.xml')
    self.assertEquals(element.Installer.class_type, installer.type)
    self.assertEquals(installer.manifest, 'http://test.com/installer.xml')

  def testSerialize(self):
    image = element.Image('http://test.com/image.png', width=100, height=100)
    s = util.serialize(image)
    k = s.keys()
    k.sort()
    # we should really only have three things to serialize
    props = s['properties']
    self.assertEquals(len(props), 3)
    self.assertEquals(props['url'], 'http://test.com/image.png')
    self.assertEquals(props['width'], 100)
    self.assertEquals(props['height'], 100)

  def testSerializeAttachment(self):
    attachment = element.Attachment(caption='My Favorite', data='SomefakeData')
    s = util.serialize(attachment)
    k = s.keys()
    k.sort()
    # we should really have two things to serialize
    props = s['properties']
    self.assertEquals(len(props), 2)
    self.assertEquals(props['caption'], 'My Favorite')
    self.assertEquals(props['data'], base64.encodestring('SomefakeData'))
    self.assertEquals(attachment.data, 'SomefakeData')

  def testSerializeLine(self):
    line = element.Line(element.Line.TYPE_H1, alignment=element.Line.ALIGN_LEFT)
    s = util.serialize(line)
    k = s.keys()
    k.sort()
    # we should really only have three things to serialize
    props = s['properties']
    self.assertEquals(len(props), 2)
    self.assertEquals(props['alignment'], 'l')
    self.assertEquals(props['lineType'], 'h1')

  def testSerializeGadget(self):
    gadget = element.Gadget('http://test.com', {'prop1': 'a', 'prop_cap': None}) 
    s = util.serialize(gadget)
    k = s.keys()
    k.sort()
    # we should really only have three things to serialize
    props = s['properties']
    self.assertEquals(len(props), 3)
    self.assertEquals(props['url'], 'http://test.com')
    self.assertEquals(props['prop1'], 'a')
    self.assertEquals(props['prop_cap'], None)

  def testGadgetElementFromJson(self):
    url = 'http://www.foo.com/gadget.xml'
    json = {
      'type': element.Gadget.class_type,
      'properties': {
        'url': url,
      }
    }
    gadget = element.Element.from_json(json)
    self.assertEquals(element.Gadget.class_type, gadget.type)
    self.assertEquals(url, gadget.url)

  def testImageElementFromJson(self):
    url = 'http://www.foo.com/image.png'
    width = '32'
    height = '32'
    attachment_id = '2'
    caption = 'Test Image'
    json = {
      'type': element.Image.class_type,
      'properties': {
        'url': url,
        'width': width,
        'height': height,
        'attachmentId': attachment_id,
        'caption': caption,
      }
    }
    image = element.Element.from_json(json)
    self.assertEquals(element.Image.class_type, image.type)
    self.assertEquals(url, image.url)
    self.assertEquals(width, image.width)
    self.assertEquals(height, image.height)
    self.assertEquals(attachment_id, image.attachmentId)
    self.assertEquals(caption, image.caption)

  def testAttachmentElementFromJson(self):
    caption = 'fake caption'
    data = 'fake data'
    mime_type = 'fake mime'
    attachment_id = 'fake id'
    attachment_url = 'fake URL'
    json = {
      'type': element.Attachment.class_type,
      'properties': {
        'caption': caption,
        'data': data,
        'mimeType': mime_type,
        'attachmentId': attachment_id,
        'attachmentUrl': attachment_url,
      }
    }
    attachment = element.Element.from_json(json)
    self.assertEquals(element.Attachment.class_type, attachment.type)
    self.assertEquals(caption, attachment.caption)
    self.assertEquals(data, attachment.data)
    self.assertEquals(mime_type, attachment.mimeType)
    self.assertEquals(attachment_id, attachment.attachmentId)
    self.assertEquals(attachment_url, attachment.attachmentUrl)

  def testFormElementFromJson(self):
    name = 'button'
    value = 'value'
    default_value = 'foo'
    json = {
      'type': element.Label.class_type,
      'properties': {
        'name': name,
        'value': value,
        'defaultValue': default_value,
      }
    }
    el = element.Element.from_json(json)
    self.assertEquals(element.Label.class_type, el.type)
    self.assertEquals(name, el.name)
    self.assertEquals(value, el.value)

  def testCanInstantiate(self):
    bag = [element.Check(name='check', value='value'),
           element.Button(name='button', value='caption'),
           element.Input(name='input', value='caption'),
           element.Label(label_for='button', caption='caption'),
           element.RadioButton(name='name', group='group'),
           element.RadioButtonGroup(name='name', value='value'),
           element.Password(name='name', value='geheim'),
           element.TextArea(name='name', value='\n\n\n'),
           element.Installer(manifest='test.com/installer.xml'),
           element.Line(line_type='type',
                        indent='3',
                        alignment='r',
                        direction='d'),
           element.Gadget(url='test.com/gadget.xml',
                          props={'key1': 'val1', 'key2': 'val2'}),
           element.Image(url='test.com/image.png', width=100, height=200),
           element.Attachment(caption='fake caption', data='fake data')]
    types_constructed = set([type(x) for x in bag])
    types_required = set(element.ALL.values())
    missing_required = types_constructed.difference(types_required)
    self.assertEquals(missing_required, set())
    missing_constructed = types_required.difference(types_constructed)
    self.assertEquals(missing_constructed, set())


if __name__ == '__main__':
  unittest.main()







# me2/eventbase.py
#
#

""" base class of all events. """

## me2 imports

from channelbase import ChannelBase
from me2.utils.lazydict import LazyDict
from me2.utils.generic import splittxt, stripped
from errors import NoSuchUser
from me2.utils.opts import makeeventopts
from me2.lib.config import Config

## simplejson imports

from me2.contrib.simplejson import dumps, loads

## basic imports

from  xml.sax.saxutils import unescape
import copy
import logging
import Queue
import types
import socket
import threading

## defines

cpy = copy.deepcopy

## classes

class EventBase(LazyDict):

    """ basic event class. """

    def __init__(self, input=None, bot=None):
        LazyDict.__init__(self)
        if bot: self.bot = bot
        if input: self.copyin(input)
        self.result = []
        self.queues = []
        self.inqueue = Queue.Queue()
        self.outqueue = Queue.Queue()
        self.bottype = "botbase"
        self.closequeue = True
        self.ttl = 1
        self.how = "normal"
        self.chantag = None
        self.resqueue = Queue.Queue()
        self.forwarded = False
        self.iscommand = False
        self.stop = False
        self.usercmnd = ""
        self.botevent = False
        self.notask = False
        self.speed = 5

    def __deepcopy__(self, a):
        """ deepcopy an event. """
        logging.debug("eventbase - cpy - %s" % type(self))
        e = EventBase()
        e.copyin(self)
        return e

    def prepare(self, bot=None):
        """ prepare the event for dispatch. """
        self.result = []
        if bot: self.bot = bot
        assert(self.bot)
        self.origin = self.channel
        self.origtxt = self.txt
        self.makeargs()
        logging.debug("%s - prepared event - %s" % (self.auth, self.cbtype))

    def bind(self, bot=None, user=None, chan=None):
        """ bind event.bot event.user and event.chan to execute a command on it. """
        target = self.auth
        bot = bot or self.bot
        assert bot
        logging.debug("eventbase - binding user - %s" % str(self.user))
        if not self.user and target:
            cfg = Config()
            if cfg.auto_register: 
                bot.users.addguest(target)
            self.user = user or bot.users.getuser(target)
        logging.debug("eventbase - binding channel - %s" % str(self.chan))
        if not self.chan:
            if chan: self.chan = chan
            elif self.channel: self.chan = ChannelBase(self.channel, bot.botname)
            elif self.userhost: self.chan = ChannelBase(self.userhost, bot.botname)
            #if bot.isgae and bot.type == "web": self.chan.gae_create()
        if not self.user: logging.info("eventbase - setting nodispatch") ; self.nodispatch = True
        self.prepare(bot)
        return self

    def parse(self, event, *args, **kwargs):
        """ overload this. """
        self.bot = event.bot
        self.origin = event.origin
        self.ruserhost = self.origin
        self.userhost = self.origin
        self.channel = event.channel
        self.auth = stripped(self.userhost)

    def copyin(self, eventin):
        """ copy in an event. """
        if not eventin:
            logging.error("no event given in copyin")
            return self
        self.update(eventin)
        self.result = []
        if eventin.has_key("sock"): self.sock = eventin['sock']
        if eventin.has_key("chan") and eventin['chan']: self.chan = eventin['chan']
        if eventin.has_key("user"): self.user = eventin['user']
        if eventin.has_key('queues'):
            if eventin['queues']: self.queues = eventin['queues']
        if eventin.has_key("inqueue"): self.inqueue = eventin['inqueue']
        if eventin.has_key("outqueue"): self.outqueue = eventin['outqueue']
        return self

    def reply(self, txt, result=[], event=None, origin="", dot=u", ", nr=375, extend=0, *args, **kwargs):
        """ reply to this event """
        if self.checkqueues(result): return
        if self.silent:
            self.msg = True
            self.bot.say(self.nick, txt, result, self.userhost, extend=extend, event=self, *args, **kwargs)
        elif self.isdcc: self.bot.say(self.sock, txt, result, self.userhost, extend=extend, event=self, *args, **kwargs)
        else: self.bot.say(self.channel, txt, result, self.userhost, extend=extend, event=self, *args, **kwargs)
        self.result.append(txt)
        return self

    def missing(self, txt):
        """ display missing arguments. """
        self.reply("%s %s" % (self.usercmnd, txt)) 
        return self

    def done(self):
        """ tell the user we are done. """
        self.reply('<b>done</b> - %s' % self.txt)
        return self

    def leave(self):
        self.ttl -= 1
        if self.ttl <= 0 : self.status = "done"
        logging.info("======== STOP handling event ========")

    def makeoptions(self):
        self.options = makeeventopts(self.txt)
        if not self.options: return
        logging.debug("eventbase - options - %s" % unicode(self.options))
        self.txt = ' '.join(self.options.args)
        self.makeargs()

    def makeargs(self):
        """ make arguments and rest attributes from self.txt. """
        if not self.txt:
            self.args = []
            self.rest = ""
        else:
            args = self.txt.split()
            self.chantag = args[0]
            if len(args) > 1:
                self.args = args[1:]
                self.rest = ' '.join(self.args)
            else:
                self.args = []
                self.rest = ""

    def checkqueues(self, resultlist):
        """ check if resultlist is to be sent to the queues. if so do it. """
        if self.queues:
            for queue in self.queues:   
                for item in resultlist:
                    if item: queue.put_nowait(item)
            for item in resultlist:
                if item: self.outqueue.put_nowait(item)      
            return True
        return False

    def makeresponse(self, txt, result, dot=u", ", *args, **kwargs):
        """ create a response from a string and result list. """
        return self.bot.makeresponse(txt, result, dot, *args, **kwargs)

    def less(self, what, nr=365):
        """ split up in parts of <nr> chars overflowing on word boundaries. """
        return self.bot.less(what, nr)

    def isremote(self):
        if self.txt: return self.txt.startswith('{"') or self.txt.startswith("{&")

    def iscmnd(self):
        """ check if event is a command. """
        if not self.txt: return
        if self.iscommand: return self.txt
        if self.isremote(): logging.info("eventbase - event is remote") ; return
        logging.debug("eventbase - trying to match %s" % self.txt)
        cc = "!"
        if not self.chan: self.chan = ChannelBase(self.channel, self.bot.botname) 
        cc = self.chan.data.cc
        if not cc:
            self.chan.data.cc = "!"
            self.chan.save()
        if not cc: cc = "!"
        if self.type == "DISPATCH": cc += "!"
        if self.txt and self.txt[0] in cc: return self.txt[1:]
        if not self.bot: logging.warn("eventbase - bot is not bind into event.") ; return False
        logging.debug("eventbase - cc for %s is %s (%s)" % (self.title or self.channel or self.userhost, cc, self.bot.nick))
        matchnick = unicode(self.bot.nick + u":")
        if self.txt.startswith(matchnick): return self.txt[len(matchnick):]
        return False



# gaelib/xmpp/event.py
#
#

""" an xmpp event. """

## me2 imports

from me2.lib.eventbase import EventBase
from me2.utils.xmpp import stripped, resource
from me2.utils.lazydict import LazyDict
from me2.lib.gae.utils.auth import checkuser
from me2.utils.generic import  strippedtxt

## xmpp import

from me2.contrib.xmlstream import XMLescape, XMLunescape

## basic imports

import cgi
import logging
import re

## XMPPEvent class

class XMPPEvent(EventBase):

    """ an XMPP event. """

    def __init__(self, bot=None): 
        EventBase.__init__(self)
        self.bottype = "xmpp"
        self.cbtype = 'MESSAGE'
        self.bot = bot

    def __deepcopy__(self, a):
        """ make a deepcopy of this XMPPEvent. """
        return XMPPEvent().copyin(self)

    def parse(self, request, response):
        """ parse incoming request/response into a XMPPEvent. """
        self.copyin(LazyDict(request.POST))
        (userhost, user, u, nick) = checkuser(response, request)
        self.userhost = self['from']
        self.origin = self.channel
        if user: self.auth = user.email()
        else: self.auth = stripped(self.userhost)
        logging.info('xmpp - auth is %s' % self.auth)
        self.resource = resource(self['from'])
        self.jid = self['from']
        self.to = stripped(self['to'])
        self.channel = stripped(self.userhost)
        self.stripped = stripped(self.userhost)
        self.nick = self.stripped.split("@")[0]
        self.origin = self.channel
        input = self.body or self.stanza
        input = input.strip()
        self.origtxt = input
        self.txt = input
        self.usercmnd = self.txt.split()[0].lower()
        logging.debug(u'xmpp - in - %s - %s' % (self.userhost, self.txt))
        return self




# me2/socklib/utils/generic.py
#
#

""" generic functions """


## me2 imports

from me2.lib.persist import Persist
from me2.utils.exception import handle_exception
from me2.utils.trace import calledfrom
from me2.utils.lazydict import LazyDict
from me2.contrib.simplejson import dumps

## generic imports

from stat import ST_UID, ST_MODE, S_IMODE

import time
import sys
import re
import getopt
import types
import os
import random
import socket
import Queue 
import logging

def jsonstring(s):
    if type(s) == types.TupleType:
        s = list(s)
    return dumps(s)

def stripident(userhost):
    """ strip ident char from userhost """
    try:
        userhost.getNode()
        return str(userhost)
    except AttributeError:  
        pass
    if not userhost:
        return None 
    if userhost[0] in "~-+^":
        userhost = userhost[1:]
    elif userhost[1] == '=':   
        userhost = userhost[2:]
    return userhost

def stripidents(ulist):
    """ strip ident char from list of userhosts """
    result = []
    for userhost in ulist:
        result.append(stripident(userhost))
    return result

def makedirs(datadir):
    if not os.path.isdir(datadir):
        os.mkdir(datadir)
    if not os.path.isdir(datadir + '/states/'):    
        os.mkdir(datadir + '/states/')    
    if not os.path.isdir(datadir + '/db/'):
        os.mkdir(datadir + '/db/')
    if not os.path.isdir(datadir + '/configs/'):
        os.mkdir(datadir + '/configs/')

def makeoptions(ievent, options={}):
    options = LazyDict(options)
    try:
        optargs = ""
        optlist = []
        if not options.has_key('--filter'):
            options['--filter'] = ""
        if not options.has_key('--to'):
            options['--to'] = None
        if not options.has_key('--chan'):
            options['--chan'] = ievent.channel
        if not options.has_key('--how'):
            options['--how'] = "msg"
        if not options.has_key('--speed'):
            options['--speed'] = str(ievent.speed)
        else:
            options['--speed'] = str(options['--speed'])
        for i, j in options.iteritems():
            if '--' in i:
                optlist.append("%s=" % i[2:])
                if j:
                    optlist.append(j)
                continue
            if '-' in i:
                if j:
                    optargs += ":%s" % i[1:]
                else:
                    optargs += i[1:]
        args = ievent.txt.split()
        try:
            (opts, rest) = getopt.getopt(args[1:], optargs, optlist)
        except AttributeError, ex:
            print "option not allowed: %s" % str(ex), ievent.txt, options
            return 0
        except getopt.GetoptError, ex:
            return 0
        if opts:
             for item in opts:
                 ievent.optionset.append(item[0])
        o = dict(options)
        o.update(dict(opts))
        try:
            filter = o['--filter']
            if filter and filter not in ievent.filter:
                ievent.filter.append(filter)
        except KeyError:
            pass
        try:
            speed = o['--speed']
            ievent.speed = int(speed)
        except KeyError:
            pass
        try:
            ievent.channel = o['--chan'] or ievent.channel
        except KeyError:
            pass
        ievent.options.update(o)
        if args:
            ievent.txt = args[0] + ' ' + ' '.join(rest)
        makeargrest(ievent)
    except Exception, ex:
        handle_exception()
        return 
    return ievent.options

def makeargrest(ievent):
    """ create ievent.args and ievent.rest .. this is needed because \
         ircevents might be created outside the parse() function """

    if not ievent.txt:
        return

    try:
        ievent.args = ievent.txt.split()[1:]
    except ValueError:
        ievent.args = []   

    try:
        cmnd, ievent.rest = ievent.txt.split(' ', 1)
    except ValueError:   
        ievent.rest = ""   

    ievent.usercmnd = ievent.txt.split()[0]

def setdefenc(encoding):
    import sys
    reload(sys)
    sys.setdefaultencoding(encoding)

def plugfile(datadir):
    return datadir + os.sep + calledfrom(sys._getframe())

def cchar(bot, ievent):
    from me2.lib.config import Config
    try:
        cchar = bot.channels[ievent.channel]['cc']
    except LookupError:
        cchar = Config()['defaultcc'] or '!'
    except TypeError:
        cchar = Config()['defaultcc'] or '!'
    return cchar

def splittxt(what, l=375):
    txtlist = []
    start = 0
    end = l
    length = len(what)
    for i in range(length/end+1):
        endword = what.find(' ', end)
        if endword == -1:
            endword = length
        res = what[start:endword]
        if res:
            txtlist.append(res)
        start = endword
        end = start + l
    return txtlist
    
class istr(str): 
    pass

def die():
    os.kill(os.getpid(), 9)

def getlistensocket(listenip):
    port = 5000
    while 1:
        time.sleep(0.01)
        try:
            port += 1
            if ':' in listenip:
                sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            else:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.setblocking(1)
            if port > 65000:
                port = 5000
            sock.bind((listenip, port))
            sock.listen(1)
            return (port, sock)
        except Exception, ex:
            pass

def checkchan(bot, item):
    chanre = re.search(' chan (\S+)', item)
    if chanre:
        chan = str(chanre.group(1))
        item = re.sub(' chan ' + re.escape(chan), '', item)
        return (chan.lower(), item)

def getwho(bot, who):
    """ get userhost from bots userhost cache """
    try:
        result = bot.userhosts[who]
        if bot.cfg['stripident']:
            logging.debug('getwho - removed ident from %s' % result)
            result = stripident(result)
        return result
    except KeyError:
        return None

def waitforuser(bot, userhost, timeout=15):
    queue = Queue.Queue()
    waitnr = bot.privwait.register(userhost, queue, timeout)
    result = queue.get()
    bot.privwait.delete(waitnr)
    return result

def getrandomnick():
    return "me2-" + str(random.randint(0, 100))

def waitforqueue(queue, timeout=10, maxitems=None):
    result = []
    while 1:
        try:
            res = queue.get(1, timeout)
        except Queue.Empty:
            break
        if not res:
            break
        result.append(res)
        if maxitems and len(result) == maxitems:
            break
    return result

def decodeperchar(txt, encoding='utf-8', what=""):

    res = []
    nogo = []

    for i in txt:
        try:
            res.append(i.decode(encoding))
        except UnicodeDecodeError:
            if i not in nogo:
                nogo.append(i)
            continue

    if nogo:
        if what:
            logging.debug("generic - %s: can't decode %s characters to %s" % (what, nogo, encoding))
        else:
            logging.debug("generic - can't decode %s characters to %s" % (nogo, encoding))

    return u"".join(res)

def toascii(what):
    what = what.encode('ascii', 'replace')
    return what

def tolatin1(what):
    what = what.encode('latin-1', 'replace')
    return what

def strippedtxt(what, allowed=[]):
    txt = []
    allowed = allowed + ['\001', '\002', '\003', '\t']
    for i in what:
        if ord(i) > 31 or (allowed and i in allowed):
            txt.append(i)
    return ''.join(txt)

def uniqlist(l):
    result = []
    for i in l:
        j = i.strip()
        if j not in result:
            result.append(j)
    return result

def fix_format(s):
    counters = {
        chr(2): 0, 
        chr(3): 0
        }
    for letter in s:
        if letter in counters:
            counters[letter] += 1
    for char in counters:
        if counters[char] % 2:
            s += char
    return s

def stripbold(s):
    s = s.replace(chr(2), '')
    s = s.replace(chr(3), '')
    return s

def jabberstrip(text, allowed=[]):
    txt = []
    allowed = allowed + ['\n', '\t']
    for i in text:
        if ord(i) > 31 or (allowed and i in allowed):
            txt.append(i)
    return ''.join(txt)

def plugnames(dirname):
    result = []
    for i in os.listdir(dirname):
        if os.path.isdir(dirname + os.sep + i):
            if os.path.exists(dirname + os.sep + i + os.sep + '__init__.py'):
                result.append(i)
        elif i.endswith('.py'):
            result.append(i[:-3])
    try:
        result.remove('__init__')
    except:
        pass
    return result

def filesize(path):
    return os.stat(path)[6]

def touch(fname):
    fd = os.open(fname, os.O_WRONLY | os.O_CREAT)
    os.close(fd)  

def stringinlist(s, l):
    for i in l:     
        if s in i:  
            return 1

def stripped(userhost):
    return userhost.split('/')[0]

def gethighest(ddir, ffile):
    highest = 0
    for i in os.listdir(ddir):
        if os.path.isdir(ddir + os.sep + i) and ffile in i:
            try:
                seqnr = i.split('.')[2]
            except IndexError:
                continue
            try:
                if int(seqnr) > highest:
                    highest = int(seqnr)
            except ValueError:
                pass
    ffile += '.' + str(highest + 1)
    return ffile

def dosed(filename, sedstring):
    try:
        f = open(filename, 'r')
    except IOError, ex:
        if 'Is a dir' in str(ex):
            return
        else:
            raise
    tmp = filename + '.tmp'
    fout = open(tmp, 'w')
    seds = sedstring.split('/')
    fr = seds[1].replace('\\', '')
    to = seds[2].replace('\\', '')
    try:
        for line in f:
            l = line.replace(fr,to)
            fout.write(l)
    finally:
        fout.flush()
        fout.close()
    try:
        os.rename(tmp, filename)
    except WindowsError:
        # no atomic operation supported on windows! error is thrown when destination exists
        os.remove(filename)
        os.rename(tmp, filename)

def convertpickle(src, target):
    import gozerbot.compat.persist
    p = gozerbot.compat.persist.Persist(src)    
    if p and p.data:
        pers = Persist(target)
        if not pers.data:
            pers.data = {}
        pers.data.update(p.data)
        try:
            pers.save()
        except TypeError:
            pers2 = Persist(target)
            if not pers2.data:
                pers2.data = {}
            for item, value in p.data.iteritems():
                pers2.data[jsonstring(item)] = value
            pers2.save()



# me2/gae/web/bot.py
#
#

""" GAE web bot. """

## me2 imports

from me2.lib.botbase import BotBase
from me2.lib.outputcache import add
from me2.utils.generic import toenc, fromenc, strippedtxt
from me2.utils.url import re_url_match
from me2.utils.timeutils import hourmin
from me2.lib.channelbase import ChannelBase

## basic imports

import logging
import re
import cgi
import urllib
import time

## WebBot class

class WebBot(BotBase):

    """ webbot just inherits from botbase for now. """

    def __init__(self, cfg=None, users=None, plugs=None, botname="gae-web", *args, **kwargs):
        BotBase.__init__(self, cfg, users, plugs, botname, *args, **kwargs)
        if self.cfg: self.cfg['type'] = u'web'
        self.isgae = True
        self.type = u"web"

    def _raw(self, txt, response=None, end=u"<br>", chan=None):
        """  put txt to the client. """
        if not txt: return 
        logging.debug("%s - web - out - %s" % (self.name, txt))
        if response: response.out.write(toenc(txt + end))
        elif chan:
            from google.appengine.api import channel
            logging.warn("%s - sending to channel %s" % (self.name, chan))
            channel.send_message(chan, txt)
                
    def outnocb(self, channel, txt, how="cache", event=None, origin=None, response=None, dotime=False, *args, **kwargs):
        txt = self.normalize(txt)
        if event and not event.how == "background": 
            logging.warn("%s - out - %s" % (self.name, txt))             
            if how == "cache": add(channel, [txt, ])
        if True:
            if "http://" in txt:
                 for item in re_url_match.findall(txt):
                     logging.debug("web - raw - found url - %s" % item)
                     url = u'<a href="%s" onclick="window.open(\'%s\'); return false;"><b>%s</b></a>' % (item, item, item)
                     try: txt = re.sub(item, url, txt)
                     except ValueError:  logging.error("web - invalid url - %s" % url)
            if dotime: txt = "[%s] %s" % (hourmin(time.time()), txt)
            if not response: self.update_web(channel, txt)
            else: self._raw(txt, response)

    def normalize(self, txt):
        #txt = cgi.escape(txt)
        txt = txt.replace("&lt;br&gt;", "<br>")
        txt = txt.replace("&lt;b&gt;", "<b>")
        txt = txt.replace("&lt;/b&gt;", "</b>")
        txt = txt.replace("&lt;i&gt;", "<i>")
        txt = txt.replace("&lt;/i&gt;", "</i>")
        txt = txt.replace("&lt;h2&gt;", "<h2>")
        txt = txt.replace("&lt;/h2&gt;", "</h2>")
        txt = txt.replace("&lt;h3&gt;", "<h3>") 
        txt = txt.replace("&lt;/h3&gt;", "</h3>")
        txt = txt.replace("&lt;li&gt;", "<li>") 
        txt = txt.replace("&lt;/li&gt;", "</li>")
        #txt = txt.replace("&lt;", "<") 
        #txt = txt.replace("&gt;", ">")
        txt = strippedtxt(txt)
        return txt

    def update_web(self, channel, txt):
        import google
        chan = ChannelBase(channel, botname="gae-web")
        logging.warn("%s - webchannels are %s" % (self.name, chan.data.webchannels))
        remove = []
        for c in chan.data.webchannels:
            try:
                self._raw(txt, chan=c)
            except google.appengine.api.channel.channel.InvalidChannelClientIdError:
                remove.append(c)
        if remove:
            for c in remove: chan.data.webchannels.remove(c)
            chan.save()



# me2/persist.py
#
#

"""
    allow data to be written to disk or BigTable in JSON format. creating 
    the persisted object restores data. 

"""

## me2 imports

from me2.utils.trace import whichmodule, calledfrom
from me2.utils.lazydict import LazyDict
from me2.utils.exception import handle_exception
from me2.utils.name import stripname
from me2.utils.locking import lockdec
from datadir import getdatadir
from cache import get, set, delete

## simplejson imports

from me2.contrib.simplejson import load, dump, loads, dumps

## basic imports

import thread
import logging
import os
import types
import copy
import sys

## try google first

try:
    from google.appengine.ext import db
    import google.appengine.api.memcache as mc
    from google.appengine.api.datastore_errors import Timeout
    logging.debug("persist - using BigTable based Persist")

    ## JSONindb class

    class JSONindb(db.Model):
        """ model to store json files in. """
        modtime = db.DateTimeProperty(auto_now=True, indexed=False)
        createtime = db.DateTimeProperty(auto_now_add=True, indexed=False)
        filename = db.StringProperty()
        content = db.TextProperty(indexed=False)

    ## Persist class

    class Persist(object):

        """ persist data attribute to database backed JSON file. """ 

        def __init__(self, filename, default={}, type="cache"):
            self.plugname = calledfrom(sys._getframe())
            if 'lib' in self.plugname: self.plugname = calledfrom(sys._getframe(1))
            try: del self.fn
            except: pass 
            self.fn = unicode(filename.strip()) # filename to save to
            self.logname = os.sep.join(self.fn.split(os.sep)[-2:])
            self.type = type
            self.counter = mcounter = mc.incr(self.fn, 1, "counters", 0)
            self.key = None
            self.obj = None
            self.init(default)

        def init(self, default={}, filename=None):
            cachetype = ""
            mcounter = mc.incr(self.fn, 1, "counters")
            logging.debug("persist - %s - %s" % (self.counter, mcounter))
            if self.type == "mem":
                tmp = get(self.fn) ; cachetype = "mem"
                if tmp: self.data = tmp ; logging.debug("persist - %s - loaded %s" % (cachetype, self.fn)) ; return
            jsontxt =  mc.get(self.fn) ; cachetype = "cache"
            if type(default) == types.DictType:
                default2 = LazyDict()
                default2.update(default)
            else: default2 = copy.deepcopy(default)
            if jsontxt is None:
                logging.debug("persist - %s - loading from db" % self.logname) 
                try:
                    try: self.obj = JSONindb.get_by_key_name(self.fn)
                    except Timeout: self.obj = JSONindb.get_by_key_name(self.fn)
                except Exception, ex:
                    # bw compat sucks
                    try: self.obj = JSONindb.get_by_key_name(self.fn)
                    except Exception, ex:
                        handle_exception()
                        self.data = default2
                        return
                if self.obj == None:
                    logging.debug("persist - %s - no entry found" % self.logname)
                    self.obj = JSONindb(key_name=self.fn)
                    self.obj.content = unicode(default)
                    self.data = default2
                    return
                jsontxt = self.obj.content
                if jsontxt: mc.set(self.fn, jsontxt)
                logging.debug('persist - jsontxt is %s' % jsontxt)
                cachetype = "file"
            else: cachetype = "cache"
            logging.debug("persist - %s - loaded %s" % (cachetype, self.fn))
            self.data = loads(jsontxt)
            if type(self.data) == types.DictType:
                d = LazyDict()
                d.update(self.data)
                self.data = d
            cfrom = whichmodule()
            if 'me2' in cfrom: 
                cfrom = whichmodule(2)
                if 'me2' in cfrom: cfrom = whichmodule(3)
            cfrom = whichmodule(2)
            if 'me2' in cfrom: 
                cfrom = whichmodule(3)
                if 'me2' in cfrom: cfrom = whichmodule(4)
            if not 'run' in self.fn: 
                if cachetype: logging.debug("persist - %s - loaded %s (%s) - %s - %s" % (cachetype, self.logname, len(jsontxt), self.data.tojson(), cfrom))
                else: logging.debug("persist - db - loaded %s (%s) - %s - %s" % (self.logname, len(jsontxt), self.data.tojson(), cfrom))
            if self.data:
                set(self.fn, self.data)

        def sync(self):
            logging.info("persist - syncing %s" % self.fn)
            data = dumps(self.data)
            mc.set(self.fn, data)
            delete(self.fn, self.data)
            return data
     
        def save(self):
            """ save json data to database. """
            bla = dumps(self.data)
            if self.obj == None:
                self.obj = JSONindb(key_name=self.fn)
                self.obj.content = bla
            else: self.obj.content = bla
            self.obj.filename = self.fn
            from google.appengine.ext import db
            key = db.run_in_transaction(self.obj.put)
            logging.debug("persist - transaction returned %s" % key)
            mc.set(self.fn, bla)
            delete(self.fn, self.data)
            cfrom = whichmodule(0)
            if 'me2' in cfrom: 
                cfrom = whichmodule(2)
                if 'me2' in cfrom: cfrom = whichmodule(3)
            logging.warn('persist - %s - saved %s (%s)' % (cfrom, self.fn, len(bla)))

        def upgrade(self, filename):
            self.init(self.data, filename=filename)


except ImportError:

    ## file based persist

    logging.debug("using file based Persist")

    ## defines

    persistlock = thread.allocate_lock()
    persistlocked = lockdec(persistlock)

    ## imports for shell bots

    from me2.lib.cache import get, set
    import fcntl

    ## classes

    class Persist(object):

        """ persist data attribute to JSON file. """

        
        def __init__(self, filename, default=None, init=True):
            """ Persist constructor """
            self.fn = filename.strip() # filename to save to
            self.logname = os.sep.join(self.fn.split(os.sep)[-2:])
            self.lock = thread.allocate_lock() # lock used when saving)
            self.data = LazyDict() # attribute to hold the data
            if init:
                if default == None: default = LazyDict()
                self.init(default)
            self.count = 0

        def init(self, default={}, filename=None):
            """ initialize the data. """
            logging.debug('persist - reading %s' % self.fn)
            cfrom = whichmodule(2)
            if 'me2' in cfrom: 
                cfrom = whichmodule(3)
                if 'me2' in cfrom: cfrom = whichmodule(4)
            gotcache = False
            cachetype = "cache"
            try:
                data = get(self.fn)
                if not data:
                   datafile = open(self.fn, 'r')
                   data = datafile.read()
                   datafile.close()
                   cachetype = "file"
                else:
                    if type(data) == types.DictType:
                        d = LazyDict()
                        d.update(data)
                    else: d = data
                    self.data = d
                    cachetype = "mem"
                    logging.debug("persist - %s - loaded %s" % (cachetype, self.fn))
                    if not 'run' in self.fn: 
                        size = len(d)
                        logging.debug("persist - mem - loaded %s (%s) - %s - %s" % (self.logname, size, self.data.tojson(), cfrom))
                    return
            except IOError, ex:
                if not 'No such file' in str(ex):
                    logging.error('persist - failed to read %s: %s' % (self.logname, str(ex)))
                    raise
                else:
                    logging.debug("persist - %s doesn't exist yet" % self.logname)
                    return
            try:
                self.data = loads(data)
                set(self.fn, self.data)
                if type(self.data) == types.DictType:
                    d = LazyDict()
                    d.update(self.data)
                    self.data = d
                logging.debug("persist - %s - loaded %s" % (cachetype, self.fn))
                if not 'run' in self.fn: 
                    size = len(data)
                    if gotcache: logging.debug("persist - cache - loaded %s (%s) - %s - %s" % (self.logname, size, self.data.tojson(), cfrom))
                    else: logging.debug("persist - file - loaded %s (%s) - %s - %s" % (self.logname, size, self.data.tojson(), cfrom))
            except Exception, ex:
                logging.error('persist - ERROR: %s' % self.fn)
                raise

        def upgrade(self, filename):
            self.init(self.data, filename=filename)
            self.save(filename)

        def get(self):
            return loads(get(self.fn)) 

        def sync(self):
            logging.info("persist - syncing %s" % self.fn)
            set(self.fn, self.data)
            return self.data

        @persistlocked
        def save(self, filename=None):
            """ persist data attribute. """
            try:
                fn = filename or self.fn
                data = dumps(self.data)
                set(fn, self.data)
                d = []
                if fn.startswith(os.sep): d = [os.sep,]
                for p in fn.split(os.sep)[:-1]:
                    if not p: continue
                    d.append(p)
                    pp = os.sep.join(d)
                    if not os.path.isdir(pp):
                        logging.info("persist - creating %s dir" % pp)
                        os.mkdir(pp)
                tmp = fn + '.tmp' # tmp file to save to
                try: datafile = open(tmp, 'w')
                except IOError, ex:
                    logging.error("persist - can't save %s: %s" % (self.fn, str(ex)))
                    return
                fcntl.flock(datafile, fcntl.LOCK_EX)
                dump(self.data, datafile)
                fcntl.flock(datafile, fcntl.LOCK_UN)
                datafile.close()
                try: os.rename(tmp, fn)
                except OSError:
                    handle_exception()
                    os.remove(fn)
                    os.rename(tmp, fn)
                if 'lastpoll' in self.logname: logging.debug('persist - %s saved (%s)' % (self.logname, len(data)))
                else: logging.warn('persist - %s saved (%s)' % (self.logname, len(data)))
            except: handle_exception()
            finally: pass

class PlugPersist(Persist):

    """ persist plug related data. data is stored in jsondata/plugs/{plugname}/{filename}. """

    def __init__(self, filename, default=None):
        plugname = calledfrom(sys._getframe())
        Persist.__init__(self, getdatadir() + os.sep + 'plugs' + os.sep + stripname(plugname) + os.sep + stripname(filename))



##   xmlstream.py
##
##   Copyright (C) 2001 Matthew Allum
##
##   This program is free software; you can redistribute it and/or modify
##   it under the terms of the GNU Lesser General Public License as published
##   by the Free Software Foundation; either version 2, or (at your option)
##   any later version.
##
##   This program is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU Lesser General Public License for more details.


"""\
xmlstream.py provides simple functionality for implementing
XML stream based network protocols. It is used as a  base
for jabber.py.

xmlstream.py manages the network connectivity and xml parsing
of the stream. When a complete 'protocol element' ( meaning a
complete child of the xmlstreams root ) is parsed the dipatch
method is called with a 'Node' instance of this structure.
The Node class is a very simple XML DOM like class for
manipulating XML documents or 'protocol elements' in this
case.

"""

# $Id: xmlstream.py,v 1.45 2004/02/03 16:33:37 snakeru Exp $

import time
import sys
import re
import socket
import logging
from base64 import encodestring
import xml.parsers.expat
import cgi
from xml.sax.saxutils import unescape, escape

__version__ = VERSION = "0.5"

ENCODING = 'utf-8'      # Though it is uncommon, this is the only right setting.
ustr = str

BLOCK_SIZE  = 1024     ## Number of bytes to get at at time via socket
                       ## transactions

def XMLescape(txt):
    "Escape XML entities"
    #logging.debug("XMLescape - incoming - %s" % txt)
    if not txt:
        return txt
    txt = txt.replace("&", "&amp;")
    txt = txt.replace("<", "&lt;")
    txt = txt.replace(">", "&gt;")
    #txt = txt.replace('"', "'")
    return txt

def XMLunescape(txt):
    "Unescape XML entities"
    if not txt:
        return txt
    txt = txt.replace("&gt;", ">")
    txt = txt.replace("&lt;", "<")
    txt = txt.replace("&amp;", "&")
    txt = txt.replace("&apos;", "'")
    return txt

class error(object):
    def __init__(self, value):
        self.value = str(value)
    def __str__(self):
        return self.value

class Node(object):
    """A simple XML DOM like class"""
    def __init__(self, tag=None, parent=None, attrs={}, payload=[], node=None):
        if node:
            if type(node)<>type(self): node=NodeBuilder(node).getDom()
            self.name,self.namespace,self.attrs,self.data,self.kids,self.parent = \
                node.name,node.namespace,node.attrs,node.data,node.kids,node.parent
        else:
            self.name,self.namespace,self.attrs,self.data,self.kids,self.parent = 'tag','',{},[],[],None

        if tag: self.namespace, self.name = (['']+tag.split())[-2:]

        if parent: self.parent = parent

#        if self.parent and not self.namespace: self.namespace=self.parent.namespace   # Doesn't checked if this neccessary

        for attr in attrs.keys():
            self.attrs[attr]=attrs[attr]

        for i in payload:
            if type(i)==type(self): self.insertNode(i)
            else: self.insertXML(i)
#            self.insertNode(Node(node=i))     # Alternative way. Needs perfomance testing.

    def setParent(self, node):
        "Set the nodes parent node."
        self.parent = node

    def getParent(self):
        "return the nodes parent node."
        return self.parent

    def getName(self):
        "Set the nodes tag name."
        return self.name

    def setName(self,val):
        "Set the nodes tag name."
        self.name = val

    def putAttr(self, key, val):
        "Add a name/value attribute to the node."
        self.attrs[key] = val

    def getAttr(self, key):
        "Get a value for the nodes named attribute."
        try: return self.attrs[key]
        except: return None

    def getAttributes(self):
        "Get a value for the nodes named attribute."
        return self.attrs

    def putData(self, data):
        "Set the nodes textual data"
        self.data.append(data)

    def insertData(self, data):
        "Set the nodes textual data"
        self.data.append(data)

    def getData(self):
        "Return the nodes textual data"
        return ''.join(self.data)

    def getDataAsParts(self):
        "Return the node data as an array"
        return self.data

    def getNamespace(self):
        "Returns the nodes namespace."
        return self.namespace

    def setNamespace(self, namespace):
        "Set the nodes namespace."
        self.namespace = namespace

    def insertTag(self, name=None, attrs={}, payload=[], node=None):
        """ Add a child tag of name 'name' to the node.

            Returns the newly created node.
        """
        newnode = Node(tag=name, parent=self, attrs=attrs, payload=payload, node=node)
        self.kids.append(newnode)
        return newnode

    def insertNode(self, node):
        "Add a child node to the node"
        self.kids.append(node)
        return node

    def insertXML(self, xml_str):
        "Add raw xml as a child of the node"
        newnode = NodeBuilder(xml_str).getDom()
        self.kids.append(newnode)
        return newnode

    def __str__(self):
        return self._xmlnode2str()

    def _xmlnode2str(self, parent=None):
        """Returns an xml ( string ) representation of the node
         and it children"""
        s = "<" + self.name
        if self.namespace:
            if parent and parent.namespace != self.namespace:
                s = s + " xmlns = '%s' " % self.namespace
        for key in self.attrs.keys():
            val = ustr(self.attrs[key])
            s = s + " %s='%s'" % ( key, XMLescape(val) )
        s = s + ">"
        cnt = 0
        if self.kids != None:
            for a in self.kids:
                if (len(self.data)-1) >= cnt: s = s + XMLescape(self.data[cnt])
                s = s + a._xmlnode2str(parent=self)
                cnt=cnt+1
        if (len(self.data)-1) >= cnt: s = s + XMLescape(self.data[cnt])
        if not self.kids and s[-1:]=='>':
            s=s[:-1]+' />'
        else:
            s = s + "</" + self.name + ">"
        return s

    def getTag(self, name, index=None):
        """Returns a child node with tag name. Returns None
        if not found."""
        for node in self.kids:
            if node.getName() == name:
                if not index: return node
                if index is not None: index-=1
        return None

    def getTags(self, name):
        """Like getTag but returns a list with matching child nodes"""
        nodes=[]
        for node in self.kids:
            if node.getName() == name:
               nodes.append(node)
        return nodes

    def getChildren(self):
        """Returns a nodes children"""
        return self.kids

    def removeTag(self,tag):
        """Pops out specified child and returns it."""
        if type(tag)==type(self):
            try:
                self.kids.remove(tag)
                return tag
            except: return None
        for node in self.kids:
            if node.getName()==tag:
                self.kids.remove(node)
                return node

class NodeBuilder(object):
    """builds a 'minidom' from data parsed to it. Primarily for insertXML
       method of Node"""
    def __init__(self,data=None):
        self._parser = xml.parsers.expat.ParserCreate()
        self._parser.StartElementHandler  = self.unknown_starttag
        self._parser.EndElementHandler    = self.unknown_endtag
        self._parser.CharacterDataHandler = self.handle_data
        self.__depth = 0
        self._dispatch_depth = 1
        self.last_is_data = False
        self._ptr = Node()
        if data: self._parser.Parse(data,1)

    def unknown_starttag(self, tag, attrs):
        """XML Parser callback"""
        self.__depth = self.__depth + 1
        if self.__depth == self._dispatch_depth:
            self._mini_dom = Node(tag=tag, attrs=attrs)
            self._ptr = self._mini_dom
        elif self.__depth > self._dispatch_depth:
            self._ptr.kids.append(Node(tag=tag,parent=self._ptr,attrs=attrs))
            self._ptr = self._ptr.kids[-1]
        else:                           ## it the stream tag:
            if attrs.has_key('id'):
                self._incomingID = attrs['id']
        self.last_is_data = False

    def unknown_endtag(self, tag ):
        """XML Parser callback"""
        if self.__depth == self._dispatch_depth:
            self.dispatch(self._mini_dom)
        elif self.__depth > self._dispatch_depth:
            self._ptr = self._ptr.parent
        self.__depth = self.__depth - 1
        self.last_is_data = False

    def handle_data(self, data):
        """XML Parser callback"""
        if self.last_is_data:
            self._ptr.data[-1] += data
        else:
            self._ptr.data.append(data)
            self.last_is_data = True

    def dispatch(self,dom):
        pass

    def getDom(self):
        return self._mini_dom



# me2/runner.py
#
#

""" threads management to run jobs. """

## me2 imports

from me2.lib.threads import getname, start_new_thread, start_bot_command
from me2.utils.exception import handle_exception
from me2.utils.locking import locked, lockdec
from me2.utils.lockmanager import rlockmanager, lockmanager
from me2.utils.trace import callstack
from me2.lib.threadloop import RunnerLoop
from me2.lib.callbacks import callbacks

## basic imports

import Queue
import time
import thread
import random
import logging
import sys

## Runner class

class Runner(RunnerLoop):

    """
        a runner is a thread with a queue on which jobs can be pushed. 
        jobs scheduled should not take too long since only one job can 
        be executed in a Runner at the same time.

    """

    def __init__(self, name="runner"):
        RunnerLoop.__init__(self, name)
        self.working = False
        self.starttime = time.time()
        self.elapsed = self.starttime
        self.finished = time.time()

    def handle(self, descr, func, *args, **kwargs):
        """ schedule a job. """
        self.working = True
        name = getname(str(func))
        try:
            rlockmanager.acquire(getname(str(func)))
            name = getname(str(func))
            self.name = name
            logging.debug('runner - running %s: %s' % (descr, name))
            self.starttime = time.time()
            func(*args, **kwargs)
            self.finished = time.time()
            self.elapsed = self.finished - self.starttime
            if self.elapsed > 3:
                logging.debug('runner - ALERT %s %s job taking too long: %s seconds' % (descr, str(func), self.elapsed))
        except Exception, ex: handle_exception()
        finally: rlockmanager.release()
        self.working = False

## BotEventRunner class

class BotEventRunner(Runner):

    def handle(self, descr, func, bot, ievent, *args, **kwargs):
        """ schedule a bot command. """
        try:
            self.starttime = time.time()
            lockmanager.acquire(getname(str(func)))
            name = getname(str(func))
            self.name = name
            self.working = True
            logging.debug("runner - now running %s" % name)
            func(bot, ievent, *args, **kwargs)
            if ievent.closequeue and ievent.queues:
                logging.debug("closing %s queues" % len(ievent.queues))
                for queue in ievent.queues: queue.put_nowait(None)
            ievent.outqueue.put_nowait(None)
            ievent.inqueue.put_nowait(None)
            self.finished = time.time()
            self.elapsed = self.finished - self.starttime
            if self.elapsed > 3:
                logging.warn('runner - ALERT %s %s job taking too long: %s seconds' % (descr, str(func), self.elapsed))
        except Exception, ex:
            if ievent.showexception: handle_exception(ievent)
            else: handle_exception(stop=True)
        finally: lockmanager.release(getname(str(func)))
        self.working = False

## Runners class

class Runners(object):

    """ runners is a collection of runner objects. """

    def __init__(self, max=100, runnertype=Runner):
        self.max = max
        self.runners = []
        self.runnertype=runnertype

    def runnersizes(self):
        """ return sizes of runner objects. """
        result = []
        for runner in self.runners: result.append(runner.queue.qsize())
        return result

    def stop(self):
        """ stop runners. """
        for runner in self.runners: runner.stop()

    def start(self):
        """ overload this if needed. """
        pass
 
    def put(self, *data):
        """ put a job on a free runner. """
        for runner in self.runners:
            if not runner.queue.qsize():
                runner.put(*data)
                return
        runner = self.makenew()
        runner.put(*data)
         
    def running(self):
        """ return list of running jobs. """
        result = []
        for runner in self.runners:
            if runner.queue.qsize(): result.append(runner.nowrunning)
        return result

    def makenew(self):
        """ create a new runner. """
        runner = None
        for i in self.runners:
            if not i.queue.qsize(): return i
        if len(self.runners) < self.max:
            runner = self.runnertype()
            runner.start()
            self.runners.append(runner)
        else: runner = random.choice(self.runners)
        return runner

    def cleanup(self):
        """ clean up idle runners. """
        if not len(self.runners): logging.info("nothing to clean")
        for index in range(len(self.runners)-1, -1, -1):
            runner = self.runners[index]
            logging.debug("runner - cleanup %s" % runner.name)
            if not runner.queue.qsize(): runner.stop() ; del self.runners[index]
            else: logging.info("runners - %s" % runner.nowrunning)
## global runners

cmndrunner = defaultrunner = longrunner = Runners(10, BotEventRunner)

def runnercleanup(bot, event):
    logging.debug("runner sizes: %s" % str(cmndrunner.runnersizes()))
    cmndrunner.cleanup()

callbacks.add("TICK", runnercleanup)


# handler_xmpp.py
#
#

""" xmpp request handler. """

## me2 imports

from me2.utils.generic import fromenc, toenc
from me2.lib.version import getversion
from me2.utils.lazydict import LazyDict
from me2.utils.exception import handle_exception
from me2.lib.plugins import plugs
from me2.lib.boot import boot, plugin_packages
from me2.contrib.simplejson import loads

## gaelib imports

from me2.lib.gae.xmpp.bot import XMPPBot
from me2.lib.gae.xmpp.event import XMPPEvent
from me2.lib.gae.utils.auth import checkuser

## google imports

from google.appengine.api import xmpp
from google.appengine.ext import webapp
from google.appengine.ext.webapp.util import run_wsgi_app
from google.appengine.ext.webapp import template
from google.appengine.api import users as gusers
from google.appengine.ext import db
from google.appengine.ext.webapp import xmpp_handlers

## basic imports

import wsgiref.handlers
import sys
import time
import types
import logging

## boot

logging.info(getversion('GAE XMPP'))

boot()

## defines

bot = XMPPBot()

## functions

def xmppbox(response):
    response.out.write("""
          <form action="/_ah/xmpp/message/chat/" method="post">
            <div><b>enter command:</b> <input type="commit" name="body">
          </form>
          """)

## classes

class XMPPHandler(webapp.RequestHandler):

    """ relay incoming messages to the bot. """

    def post(self):
        try:
            logging.info("XMPP incoming: %s" % self.request.remote_addr)

            if not self.request.POST.has_key('from'):
                logging.debug('no from in POST: %s' % str(self.request.POST))
                return

            if not self.request.POST.has_key('to'):
                logging.debug('no to in POST: %s' % str(self.request.POST))
                return

            event = XMPPEvent(bot=bot).parse(self.request, self.response)
            bot.doevent(event)

        except Exception, ex:
            handle_exception()
            #self.send_error(500)

application = webapp.WSGIApplication([('/_ah/xmpp/message/chat/', XMPPHandler), ],
                                      debug=True)

def main():
    global application
    global bot
    run_wsgi_app(application)

if __name__ == "__main__":
    main()



# plugs/shop.py
#
#

""" maitain a shopping list (per user). """

## me2 imports

from me2.utils.generic import getwho, jsonstring
from me2.lib.users import users
from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.utils.pdol import Pdol

## basic imports

import os

## functions

def size():
    """ return number of shops entries """
    return len(shops.data)

def sayshop(bot, ievent, shoplist):
    """ output shoplist """
    if not shoplist:
        ievent.reply('nothing to shop ;]')
        return
    result = []
    teller = 0
    for i in shoplist:
        result.append('%s) %s' % (teller, i))
        teller += 1
    ievent.reply("shoplist: ", result, dot=" ")

## commands

def handle_shop(bot, ievent):
    """ shop [<item>] .. show shop list or add <item> """
    if len(ievent.args) != 0:
        handle_shop2(bot, ievent)
        return

    if ievent.user.data.shops:
        sayshop(bot, ievent, ievent.user.data.shops)
    else:
        ievent.reply("no shops")

cmnds.add('shop', handle_shop, ['USER', 'GUEST'])

def handle_shop2(bot, ievent):
    """ add items to shop list """
    if not ievent.rest:
        ievent.missing('<shopitem>')
        return
    else:
        what = ievent.rest
    if not ievent.user.data.shops:
        ievent.user.data.shops = []
    ievent.user.data.shops.append(what)
    ievent.user.save()
    ievent.reply('shop item added')

examples.add('shop', 'shop [<item>] .. show shop items or add a shop item', '1) shop 2) shop bread')

def handle_got(bot, ievent):
    """ got <listofnrs> .. remove items from shoplist """
    if len(ievent.args) == 0:
        ievent.missing('<list of nrs>')
        return
    try:
        nrs = []
        for i in ievent.args:
            nrs.append(int(i))
    except ValueError:
        ievent.reply('%s is not an integer' % i)
        return
    try:
        shop = ievent.user.data.shops
    except KeyError:
        ievent.reply('nothing to shop ;]')
        return
    if not shop:
        ievent.reply("nothing to shop ;]")
        return
    nrs.sort()
    nrs.reverse()
    teller = 0
    for i in range(len(shop)-1, -1 , -1):
        if i in nrs:
            try:
                del shop[i]
                teller += 1
            except IndexError:
                pass
    ievent.user.save()
    ievent.reply('%s shop item(s) deleted' % teller)

cmnds.add('got', handle_got, ['USER', 'GUEST'])
examples.add('got', 'got <listofnrs> .. remove item <listofnrs> from shop list','1) got 3 2) got 1 6 8')




# me2/commands.py
#
#

""" 
    the commands module provides the infrastructure to dispatch commands. 
    commands are the first word of a line. 

"""

## me2 imports

from threads import start_new_thread, start_bot_command
from me2.utils.xmpp import stripped
from me2.utils.trace import calledfrom, whichmodule
from me2.utils.exception import handle_exception
from me2.utils.lazydict import LazyDict
from errors import NoSuchCommand
from persiststate import UserState
from runner import cmndrunner
from boot import getcmndperms

## basic imports

import logging
import sys
import types
import os

## Command class

class Command(LazyDict):

    """ a command object. """

    def __init__(self, modname, cmnd, func, perms=[], threaded=False, wait=False, orig=None):
        LazyDict.__init__(self)
        if not modname: raise Exception("modname is not set - %s" % cmnd)
        self.modname = modname
        self.plugname = self.modname.split('.')[-1]
        self.cmnd = cmnd
        self.orig = orig
        self.func = func
        if type(perms) == types.StringType: perms = [perms, ]
        self.perms = perms
        self.plugin = self.plugname
        self.threaded = threaded
        self.wait = wait
        self.enable = True

class Commands(LazyDict):

    """
        the commands object holds all commands of the bot.
 
    """

    def add(self, cmnd, func, perms, threaded=False, wait=False, *args, **kwargs):
        """ add a command. """
        modname = calledfrom(sys._getframe())
        self[cmnd] = Command(modname, cmnd, func, perms, threaded, wait)
        try:
            c = cmnd.split('-')[1]
            if not self.subs: self.subs = LazyDict()
            if self.subs.has_key(c):
                if not self.subs[c]: self.subs[c] = []
                self.subs[c].append(Command(modname, c, func, perms, threaded, wait, cmnd))
            else: self.subs[c] = [Command(modname, c, func, perms, threaded, wait, cmnd), ]
        except IndexError: pass
        return self

    def dispatch(self, bot, event, wait=0):
        """ 
            dispatch an event if cmnd exists and user is allowed to exec this 
            command.

        """

        event.bind(bot)
        if event.groupchat: id = event.auth = event.userhost
        else: id = event.auth
        #if bot.cfg.auto_register: bot.users.addguest(id)
        if not event.user:
            event.user = bot.users.getuser(id)
            if event.user: event.userstate = UserState(event.user.data.name)
            else: logging.debug("failed to set user %s" % id)
        cmnd = event.usercmnd.lower()
        try:
            cmnd = event.chan.data.aliases[cmnd]
            event.txt = event.chan.data.cc + cmnd +  ' ' + ' '.join(event.txt.split()[1:])
            event.usercmnd = cmnd.split()[0]
            event.prepare()
        except (TypeError, KeyError, AttributeError): pass
        target = bot.plugs
        if target: target.reloadcheck(bot, event)
        try:
            c = self[event.usercmnd]
        except KeyError:
            if self.subs and self.subs.has_key(cmnd):
                if len(self.subs[cmnd]) == 1: c = self.subs[cmnd][0]
                else:
                    event.reply("use one of ", [c.orig for c in self.subs[cmnd]])
                    return
            else:
                raise NoSuchCommand(cmnd)

        ## core business
        if bot.cmndperms and bot.cmndperms[c.cmnd]: perms = bot.cmndperms[c.cmnd]
        else: perms = c.perms
        if bot.allowall: return self.doit(bot, event, c, wait=wait)
        elif event.chan and event.chan.data.allowcommands and event.usercmnd in event.chan.data.allowcommands: 
            if not 'OPER' in perms:  return self.doit(bot, event, c, wait=wait)
            else: logging.warn("commands - %s is not in allowlist" % c)
        elif not bot.users or bot.users.allowed(id, perms, bot=bot): return self.doit(bot, event, c, wait=wait)
        elif bot.users.allowed(id, perms, bot=bot): return self.doit(bot, event, c, wait=wait)
        return event

    def doit(self, bot, event, target, wait=0):
        """ do the dispatching. """
        if not target.enable: return
        if target.modname in event.chan.data.denyplug:
             logging.warn("commands - %s is denied in channel %s - %s" % (target.plugname, event.channel, event.userhost))
             return
        id = event.auth or event.userhost
        event.iscommand = True
        logging.warning('commands - dispatching %s for %s' % (event.usercmnd, id))
        try:
            if bot.isgae or wait:
                if bot.isgae and not event.notask and (target.threaded or event.threaded):
                    logging.warn("commands - LAUNCHING AS TASK")
                    from me2.lib.gae.tasks import start_botevent
                    event.txt = event.origtxt
                    start_botevent(bot, event, event.speed)
                else:
                    target.func(bot, event)
                    if event.closequeue and event.queues:
                        for q in event.queues:
                            q.put_nowait(None)
                        event.outqueue.put_nowait(None)
            else:
                if target.threaded:
                    logging.warning("commands - launching thread for %s" % event.usercmnd)
                    thread = start_bot_command(target.func, (bot, event))
                    if bot.isgae and event.closequeue:
                        if event.queues:
                            for q in event.queues: q.put_nowait(None)
                        event.outqueue.put_nowait(None)
                else: cmndrunner.put(target.modname, target.func, bot, event)
        except Exception, ex:
            logging.error('commands - %s - error executing %s' % (whichmodule(), str(target.func)))
            raise
        return event

    def unload(self, modname):
        """ remove modname registered commands from store. """
        delete = []
        for name, cmnd in self.iteritems():
            if not cmnd: continue
            if cmnd.modname == modname: delete.append(cmnd)
        for cmnd in delete: cmnd.enable = False
        return self

    def apropos(self, search):
        """ search existing commands for search term. """
        result = []
        for name, cmnd in self.iteritems():
            if search in name: result.append(name)
        return result

    def perms(self, cmnd):
        """ show what permissions are needed to execute cmnd. """
        try: return self[cmnd].perms
        except KeyError: return []

    def whereis(self, cmnd):
        """ return plugin name in which command is implemented. """
        try: return self[cmnd].plugname
        except KeyError: return ""

    def gethelp(self, cmnd):
        """ get the docstring of a command. used for help. """
        try: return self[cmnd].func.__doc__
        except KeyError: pass

## global commands

cmnds = Commands()



#!/usr/bin/python
#
# Copyright (C) 2009 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the generic robot classes.

This module provides the Robot class and RobotListener interface,
as well as some helper functions for web requests and responses.
"""

import base64
import logging
import sys

try:
  __import__('google3') # setup internal test environment
except ImportError:
  pass

import simplejson

import blip
import events
import ops
import util
import wavelet
import errors

# We only import oauth when we need it
oauth = None

DEFAULT_PROFILE_URL = (
    'http://code.google.com/apis/wave/extensions/robots/python-tutorial.html')

class Robot(object):
  """Robot metadata class.

  This class holds on to basic robot information like the name and profile.
  It also maintains the list of event handlers and cron jobs and
  dispatches events to the appropriate handlers.
  """

  def __init__(self, name, image_url='', profile_url=DEFAULT_PROFILE_URL):
    """Initializes self with robot information.

    Args:
      name: The name of the robot
      image_url: (optional) url of an image that should be used as the avatar
          for this robot.
      profile_url: (optional) url of a webpage with more information about
          this robot.
    """
    self._handlers = {}
    self._name = name
    self._verification_token = None
    self._st = None
    self._consumer_key = None
    self._consumer_secret = None
    self._server_rpc_base = None
    self._profile_handler = None
    self._image_url = image_url
    self._profile_url = profile_url
    self._capability_hash = 0

  @property
  def name(self):
    """Returns the name of the robot."""
    return self._name

  @property
  def image_url(self):
    """Returns the URL of the avatar image."""
    return self._image_url

  @property
  def profile_url(self):
    """Returns the URL of an info page for the robot."""
    return self._profile_url


  def http_post(self, url, data, headers):
    """Execute an http post.

    Monkey patch this method to use something other than
    the default urllib.

    Args:
        url: to post to
        body: post body
        headers: extra headers to pass along
    Returns:
        response_code, returned_page
    """
    import urllib2
    req = urllib2.Request(url,
                          data=data,
                          headers=headers)
    try:
      f = urllib2.urlopen(req)
      return f.code, f.read()
    except urllib2.URLError, e:
      return e.code, e.read()

  def get_verification_token_info(self):
    """Returns the verification token and ST parameter."""
    return self._verification_token, self._st

  def capabilities_hash(self):
    """Return the capabilities hash as a hex string."""
    return hex(self._capability_hash)

  def register_handler(self, event_class, handler, context=None, filter=None):
    """Registers a handler on a specific event type.

    Multiple handlers may be registered on a single event type and are
    guaranteed to be called in order of registration.

    The handler takes two arguments, the event object and the corresponding
    wavelet.

    Args:
      event_class: An event to listen for from the classes defined in the
          events module.

      handler: A function handler which takes two arguments, the wavelet for
          the event and the event object.

      context: The context to provide for this handler.

      filter: Depending on the event, a filter can be specified that restricts
          for which values the event handler will be called from the server.
          Valuable to restrict the amount of traffic send to the robot.
    """
    payload = (handler, event_class, context, filter)
    self._handlers.setdefault(event_class.type, []).append(payload)
    if type(context) == list:
      context = ','.join(context)
    self._capability_hash = (self._capability_hash * 13 +
                             hash(event_class.type) +
                             hash(context) +
                             hash(filter)) & 0xfffffff

  def set_verification_token_info(self, token, st=None):
    """Set the verification token used in the ownership verification.

    /wave/robot/register starts this process up and will produce this token.

    Args:
      token: the token provided by /wave/robot/register.

      st: optional parameter to verify the request for the token came from
          the wave server.
    """
    self._verification_token = token
    self._st = st

  def setup_oauth(self, consumer_key, consumer_secret,
                 server_rpc_base='http://gmodules.com/api/rpc'):
    """Configure this robot to use the oauth'd json rpc.

    Args:
      consumer_key: consumer key received from the verification process.

      consumer_secret: secret received from the verification process.

      server_rpc_base: url of the rpc gateway to use. Specify None for default.
          For wave preview, http://gmodules.com/api/rpc should be used.
          For wave sandbox, http://sandbox.gmodules.com/api/rpc should be used.
    """
    # Import oauth inline and using __import__ for pyexe compatibility
    # when oauth is not installed.
    global oauth
    __import__('waveapi.oauth')
    oauth = sys.modules['waveapi.oauth']

    self._server_rpc_base = server_rpc_base
    self._consumer_key = consumer_key
    self._consumer_secret = consumer_secret
    self._oauth_signature_method = oauth.OAuthSignatureMethod_HMAC_SHA1()
    self._oauth_consumer = oauth.OAuthConsumer(self._consumer_key,
                                               self._consumer_secret)

  def register_profile_handler(self, handler):
    """Sets the profile handler for this robot.

    The profile handler will be called when a profile is needed. The handler
    gets passed the name for which a profile is needed or None for the
    robot itself. A dictionary with keys for name, imageUrl and
    profileUrl should be returned.
    """
    self._profile_handler = handler

  def _hash(self, value):
    """return b64encoded sha1 hash of value."""
    try:
      hashlib = __import__('hashlib') # 2.5
      hashed = hashlib.sha1(value)
    except ImportError:
      import sha # deprecated
      hashed = sha.sha(value)
    return base64.b64encode(hashed.digest())


  def make_rpc(self, operations):
    """Make an rpc call, submitting the specified operations."""

    if not oauth or not self._oauth_consumer.key:
      raise errors.Error('OAuth has not been configured')
    if (not type(operations) == list and
        not isinstance(operations, ops.OperationQueue)):
      operations = [operations]

    rpcs = [op.serialize(method_prefix='wave') for op in operations]

    post_body = simplejson.dumps(rpcs)
    body_hash = self._hash(post_body)
    params = {
      'oauth_consumer_key': 'google.com:' + self._oauth_consumer.key,
      'oauth_timestamp': oauth.generate_timestamp(),
      'oauth_nonce': oauth.generate_nonce(),
      'oauth_version': oauth.OAuthRequest.version,
      'oauth_body_hash': body_hash,
    }
    oauth_request = oauth.OAuthRequest.from_request('POST',
                                                    self._server_rpc_base,
                                                    parameters=params)
    oauth_request.sign_request(self._oauth_signature_method,
                               self._oauth_consumer,
                               None)
    code, content = self.http_post(
        url=oauth_request.to_url(),
        data=post_body,
        headers={'Content-Type': 'application/json'})
    logging.info('Active URL: %s'  % oauth_request.to_url())
    logging.info('Active Outgoing: %s' % post_body)
    if code != 200:
      logging.info(oauth_request.to_url())
      logging.info(content)
      raise IOError('HttpError ' + str(code))
    return simplejson.loads(content)
  
  def _first_rpc_result(self, result):
    """result is returned from make_rpc. Get the first data record
    or throw an exception if it was an error."""
    if type(result) == list:
      result = result[0]
    error = result.get('error')
    if error:
      raise errors.Error('RPC Error' + str(error['code'])
          + ': ' + error['message'])
    data = result.get('data')
    if data:
      return data
    raise errors.Error('RPC Error: No data record.')

  def capabilities_xml(self):
    """Return this robot's capabilities as an XML string."""
    lines = []
    for capability, payloads in self._handlers.items():
      for payload in payloads:
        handler, event_class, context, filter = payload
        line = '  <w:capability name="%s"' % capability
        if context: 
          if type(context) == list:
            context = ','.join(context)
          line += ' context="%s"' % context
        if filter:
          line += ' filter="%s"' % filter
        line += '/>\n'
        lines.append(line)
    if self._consumer_key:
      oauth_tag = '<w:consumer_key>%s</w:consumer_key>\n' % self._consumer_key
    else:
      oauth_tag = ''
    return ('<?xml version="1.0"?>\n'
            '<w:robot xmlns:w="http://wave.google.com/extensions/robots/1.0">\n'
            '<w:version>%s</w:version>\n'
            '%s'
            '<w:protocolversion>%s</w:protocolversion>\n'
            '<w:capabilities>\n'
            '%s'
            '</w:capabilities>\n'
            '</w:robot>\n') % (self.capabilities_hash(),
                               oauth_tag,
                               ops.PROTOCOL_VERSION,
                               '\n'.join(lines))

  def profile_json(self, name=None):
    """Returns a JSON representation of the profile.

    This method is called both for the basic profile of the robot and to
    get a proxying for profile, in which case name is set. By default
    the information supplied at registration is returned.

    Use register_profile_handler to override this default behavior.
    """
    if self._profile_handler:
      data = self._profile_handler(name)
    else:
      data = {'name': self.name,
              'imageUrl': self.image_url,
              'profileUrl': self.profile_url}
    return simplejson.dumps(data)

  def _wavelet_from_json(self, json, pending_ops):
    """Construct a wavelet from the passed json.

    The json should either contain a wavelet and a blips record that
    define those respective object. The returned wavelet
    will be constructed using the passed pending_ops
    OperationQueue.
    Alternatively the json can be the result of a previous
    wavelet.serialize() call. In that case the blips will
    be contaned in the wavelet record.
    """
    if isinstance(json, basestring):
      json = simplejson.loads(json)

    blips = {}
    for blip_id, raw_blip_data in json['blips'].items():
      blips[blip_id] = blip.Blip(raw_blip_data, blips, pending_ops)

    if 'wavelet' in json:
      raw_wavelet_data = json['wavelet']
    elif 'waveletData' in json:
      raw_wavelet_data = json['waveletData']
    else:
      raw_wavelet_data = json
    wavelet_blips = {}
    wavelet_id = raw_wavelet_data['waveletId']
    wave_id = raw_wavelet_data['waveId']
    for blip_id, instance in blips.items():
      if instance.wavelet_id == wavelet_id and instance.wave_id == wave_id:
        wavelet_blips[blip_id] = instance
    result = wavelet.Wavelet(raw_wavelet_data, wavelet_blips, self, pending_ops)
    robot_address = json.get('robotAddress')
    if robot_address:
      result.robot_address = robot_address
    return result

  def process_events(self, json):
    """Process an incoming set of events encoded as json."""
    parsed = simplejson.loads(json)
    pending_ops = ops.OperationQueue()
    event_wavelet = self._wavelet_from_json(parsed, pending_ops)

    for event_data in parsed['events']:
      for payload in self._handlers.get(event_data['type'], []):
        handler, event_class, context, filter = payload
        event = event_class(event_data, event_wavelet)
        handler(event, event_wavelet)

    pending_ops.set_capability_hash(self.capabilities_hash())
    return simplejson.dumps(pending_ops.serialize())

  def new_wave(self, domain, participants=None, message='', proxy_for_id=None,
               submit=False):
    """Create a new wave with the initial participants on it.

    A new wave is returned with its own operation queue. It the
    responsibility of the caller to make sure this wave gets
    submitted to the server, either by calling robot.submit() or
    by calling .submit_with() on the returned wave.

    Args:
      domain: the domain to create the wavelet on. This should
          in general correspond to the domain of the incoming
          wavelet. (wavelet.domain). Exceptions are situations
          where the robot is calling new_wave outside of an
          event or when the server is handling multiple domains.

      participants: initial participants on the wave. The robot
          as the creator of the wave is always added.

      message: a string that will be passed back to the robot
          when the WAVELET_CREATOR event is fired. This is a
          lightweight way to pass around state.

      submit: if true, use the active gateway to make a round
          trip to the server. This will return immediately an
          actual waveid/waveletid and blipId for the root blip.

    """
    operation_queue = ops.OperationQueue(proxy_for_id)
    if not isinstance(message, basestring):
      message = simplejson.dumps(message)

    blip_data, wavelet_data = operation_queue.robot_create_wavelet(
        domain=domain,
        participants=participants,
        message=message)

    blips = {}
    root_blip = blip.Blip(blip_data, blips, operation_queue)
    blips[root_blip.blip_id] = root_blip
    created = wavelet.Wavelet(wavelet_data,
                              blips=blips,
                              robot=self,
                              operation_queue=operation_queue)
    if submit:
      result = self._first_rpc_result(self.submit(created))
      if type(result) == list:
        result = result[0]
      # Currently, data is sometimes wrapped in an outer 'data'
      # Remove these 2 lines when that is no longer an issue.
      if 'data' in result and len(result) == 2:
        result = result['data']
      if 'blipId' in result:
        blip_data['blipId'] = result['blipId']
        wavelet_data['rootBlipId'] = result['blipId']
      for field in 'waveId', 'waveletId':
        if field in result:
          wavelet_data[field] = result[field]
          blip_data[field] = result[field]
      blips = {}
      root_blip = blip.Blip(blip_data, blips, operation_queue)
      blips[root_blip.blip_id] = root_blip
      created = wavelet.Wavelet(wavelet_data,
                                blips=blips,
                                robot=self,
                                operation_queue=operation_queue)

    return created

  def fetch_wavelet(self, wave_id, wavelet_id, proxy_for_id=None):
    """Use the REST interface to fetch a wave and return it.

    The returned wavelet contains a snapshot of the state of the
    wavelet at that point. It can be used to modify the wavelet,
    but the wavelet might change in between, so treat carefully.

    Also note that the wavelet returned has its own operation
    queue. It the responsibility of the caller to make sure this
    wavelet gets submited to the server, either by calling
    robot.submit() or by calling .submit_with() on the returned
    wavelet.
    """
    operation_queue = ops.OperationQueue(proxy_for_id)
    operation_queue.robot_fetch_wave(wave_id, wavelet_id)
    result = self._first_rpc_result(self.make_rpc(operation_queue))
    return self._wavelet_from_json(result, ops.OperationQueue(proxy_for_id))

  def blind_wavelet(self, json, proxy_for_id=None):
    """Construct a blind wave from a json string.

    Call this method if you have a snapshot of a wave that you
    want to operate on outside of an event. Since the wave might
    have changed since you last saw it, you should take care to
    submit operations that are as safe as possible.

    Args:
      json: a json object or string containing at least a key
        wavelet defining the wavelet and a key blips defining the
        blips in the view.

      proxy_for_id: the proxying information that will be set on the wavelet's
        operation queue.

    Returns:
      A new wavelet with its own operation queue. It the
      responsibility of the caller to make sure this wavelet gets
      submited to the server, either by calling robot.submit() or
      by calling .submit_with() on the returned wavelet.
    """
    return self._wavelet_from_json(json, ops.OperationQueue(proxy_for_id))

  def submit(self, wavelet_to_submit):
    """Submit the pending operations associated with wavelet_to_submit.

    Typically the wavelet will be the result of fetch_wavelet, blind_wavelet
    or new_wave.
    """
    pending = wavelet_to_submit.get_operation_queue()
    res = self.make_rpc(pending)
    pending.clear()
    logging.info('submit returned:%s', res)
    return res



# handler_web.py
#
#

""" web request handler. """

import time
import logging

## me2 imports

from me2.lib.version import getversion
from me2.utils.exception import handle_exception
from me2.lib.channelbase import ChannelBase

## gaelib import

from me2.lib.gae.utils.auth import finduser
from me2.lib.gae.utils.web import start, closer, loginurl, logouturl, login

## google imports

from webapp2 import RequestHandler, Route, WSGIApplication
from google.appengine.ext.webapp import template
from google.appengine.api import channel

## simplejson import

from me2.contrib.simplejson import loads

## basic imports

import sys
import time
import types
import os
import logging
import google

## init

logging.info(getversion('GAE WEB'))

## classes

class HomePageHandler(RequestHandler):

    """ the bots web command dispatcher. """


    def options(self):
        self.response.headers.add_header("Allow: *")
        
    def get(self):
        """ show basic page. """

        logging.warn("web_handler - in")
        try:
            user = finduser()
            if not user:
                login(self.response, {'appname': 'BOTME2' , 'who': 'not logged in yet', 'loginurl': 'not logged in', 'logouturl': 'BOTME2', 'onload': 'consoleinit();'})
            else:
                logout = logouturl(self.request, self.response)
                channel = ChannelBase(user, botname="gae-web")
                (chan, token) = channel.gae_create()
                logging.warn("%s - %s" % (chan, token))
                channel.save()
                start(self.response, {'appname': 'BOTME2' , 'who': user, 'loginurl': 'logged in', 'logouturl': logout, 'onload': 'consoleinit();', "token": token, 'webchan': chan})
        except google.appengine.runtime.DeadlineExceededError:
            self.response.out.write("DeadLineExceededError .. this request took too long to finish.")
        except Exception, ex:
            self.response.out.write("An exception occured: %s" % str(ex))
            handle_exception()
        logging.warn("web_handler - out")
        
## the application 

application = WSGIApplication([('/', HomePageHandler)],
                               debug=True)

## main

def main():
    global application
    application.run()

if __name__ == "__main__":
    main()



# me2/eventhandler.py
#
#

""" event handler. use to dispatch function in main loop. """

## me2 imports

from me2.utils.exception import handle_exception
from me2.utils.locking import lockdec
from threads import start_new_thread

## basic imports

import Queue
import thread

## locks

handlerlock = thread.allocate_lock()
locked = lockdec(handlerlock)

## classes

class EventHandler(object):

    """
        events are handled in 11 queues with different priorities:
        queue0 is tried first queue10 last.

    """

    def __init__(self):
        self.sortedlist = []
        self.queues = {}
        for i in range(11):
            self.queues[i] = Queue.Queue()
            self.sortedlist.append(i)
        self.sortedlist.sort()
        self.go = Queue.Queue()
        self.stopped = False
        self.running = False
        self.nooutput = False

    def start(self):
        """ start the eventhandler thread. """
        self.stopped = False
        if not self.running:
            start_new_thread(self.handleloop, ())
            self.running = True

    def stop(self):
        """ stop the eventhandler thread. """
        self.running = False
        self.stopped = True
        self.go.put('Yihaaa')

    def put(self, speed, func, *args, **kwargs):
        """ put item on the queue. """
        self.queues[10-speed].put_nowait((func, args, kwargs))
        self.go.put('go')

    def getready(self):
        """ check queues from available functions to execute. """
        ready = []
        for i in self.sortedlist:
            if self.queues[i].qsize():
                ready.append(i)
                break
        return ready

    def handle_one(self):
        """ do 1 loop over ready queues. """
        ready = self.getready()
        for i in ready: self.dispatch(self.queues[i])

    def handleloop(self):
        """ thread that polls the queues for items to dispatch. """
        logging.debug('eventhandler - starting handle thread')
        while not self.stopped:
            self.go.get()
            self.handle_one()
        logging.debug('eventhandler - stopping %s' % str(self))

    def dispatch(self, queue):
        """ dispatch functions from provided queue. """
        try: todo = queue.get_nowait()
        except Queue.Empty: return
        try:
            (func, args, kwargs) = todo
            func(*args, **kwargs)
        except ValueError:
            try:
                (func, args) = todo
                func(*args)
            except ValueError:
                (func, ) = todo
                func()
        except: handle_exception()


## handler to use in main prog

mainhandler = EventHandler()



# me2/socklib/xmpp/errors.py
#
#

""" xmpp error codes. """

xmpperrors = {'400': 'Bad Request',
              '401': 'Unauthorized',
              '402': 'Payment Required',
              '403': 'Forbidden',
              '404': 'Not Found',
              '405': 'Not Allowed',
              '406': 'Not Acceptable',
              '407': 'Registration Required',
              '408': 'Request Timeout',
              '409': 'Conflict',
              '500': 'Internal Server Error',
              '501': 'Not Implemented',
              '502': 'Remove Server Error',
              '503': 'Service Unavailable',
              '504': 'Remove Server Timeout',
              '510': 'Disconnected'
             }



# me2.plugs.socket/restserver.py
#
#

## me2 imports

from me2.lib.callbacks import callbacks
from me2.utils.url import posturl, getpostdata
from me2.lib.persistconfig import PersistConfig
from me2.lib.commands import cmnds
from me2.lib.socklib.rest.server import RestServer, RestRequestHandler
from me2.lib.eventbase import EventBase
from me2.utils.exception import handle_exception
from me2.lib.examples import examples

## simplejson imports

from me2.contrib.simplejson import dumps

## basic imports

import socket
import re
import logging

## defines

enable = False

try:
    cfg = PersistConfig()
    cfg.define('enable', 0)
    cfg.define('host' , socket.gethostbyname(socket.getfqdn()))
    cfg.define('name' , socket.getfqdn())
    cfg.define('port' , 11111)
    cfg.define('disable', [])
    hp = "%s:%s" % (cfg.get('host'), cfg.get('port'))
    url = "http://%s" % hp
    if cfg.enable:
         enable = True
except AttributeError:
    # we are on the GAE
    enable = False

## server part

server = None

## functions

def startserver(force=False):
    if not enable:
        logging.debug("rest server is disabled")
        return

    global server 
    if server and not force:
        logging.debug("REST server is already running. ")
        return server

    try:
        server = RestServer((cfg.get('host'), cfg.get('port')), RestRequestHandler)

        if server:
            server.start()
            logging.warn('restserver - running at %s:%s' % (cfg.get('host'), cfg.get('port')))

            for mount in cfg.get('disable'):
                server.disable(mount)
        else:
            logging.error('restserver - failed to start server at %s:%s' % (cfg.get('host'), cfg.get('port')))

    except socket.error, ex:
        logging.warn('restserver - start - socket error: %s' % str(ex))

    except Exception, ex:
        handle_exception()

    return server

def stopserver():

    try:
        if not server:
            logging.debug('restserver - server is already stopped')
            return

        server.shutdown()

    except Exception, ex:
        handle_exception()
        pass

## plugin init

def init():

    if cfg['enable']:
        startserver()

def shutdown():

    if cfg['enable']:
        stopserver()

def handle_rest_start(bot, event):
    cfg['enable'] = 1
    cfg.save()
    startserver()
    event.done()

cmnds.add('rest-start', handle_rest_start, 'OPER')
examples.add('rest-start', 'start the REST server', 'rest-start')

def handle_rest_stop(bot, event):
    cfg['enable'] = 0
    cfg.save()
    stopserver()
    event.done()

cmnds.add('rest-stop', handle_rest_stop, 'OPER')
examples.add('rest-stop', 'stop the REST server', 'rest-stop')



# handler_wave.py
#
#

""" this handler handles all the wave jsonrpc requests. """

## me2 imports

from me2.lib.version import getversion
from me2.lib.errors import NoSuchCommand
from me2.lib.boot import boot

## gaelib imports

from me2.lib.gae.wave.bot import WaveBot

## basic imports

import logging
import os

## defines

logging.info(getversion('GAE WAVE'))
boot()

# the bot

bot = WaveBot(domain="googlewave.com")

def main():
    bot.run()

if __name__ == "__main__":
    main()



# plugs/more.py
#
#

""" access the output cache. """

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.lib.less import outcache

## basic imports

import logging

## more command

def handle_more(bot, ievent):
    """ pop message from the output cache. """
    if ievent.msg: target = ievent.nick
    else: target = ievent.channel
    try: txt, size = outcache.more(u"%s-%s" % (bot.name, target))
    except IndexError: txt = None 
    if not txt:
        ievent.reply('no more data available for %s' % target)
        return
    if size: txt += "<b> - %s more</b>" % str(size)
    bot.outnocb(target, txt, response=ievent.response)
    bot.outmonitor(ievent.origin or ievent.userhost, ievent.channel, txt)

cmnds.add('more', handle_more, ['USER', 'GUEST'])
examples.add('more', 'return txt from output cache', 'more')



# handler_dispatch.py
#
#

""" me2 dispatch handler.  dispatches remote commands.  """

## boot

from me2.lib.boot import boot
boot("gozerdata")

## me2 imports

from me2.utils.generic import fromenc, toenc
from me2.lib.version import getversion
from me2.utils.xmpp import stripped
from me2.utils.url import getpostdata, useragent
from me2.lib.plugins import plugs
from me2.lib.persist import Persist
from me2.utils.exception import handle_exception, exceptionmsg
from me2.lib.fleet import fleet
from me2.lib.errors import NoSuchCommand
from me2.lib.gae.utils.web import loginurl

## gaelib imports

from me2.lib.gae.web.bot import WebBot
from me2.lib.gae.web.event import WebEvent
from me2.lib.gae.utils.auth import checkuser

## google imports

from webapp2 import RequestHandler, Route, WSGIApplication
from google.appengine.ext.webapp import template
from google.appengine.api import users as gusers

## simplejson import

from me2.contrib.simplejson import loads

## basic imports

import sys
import time
import types
import os
import logging
import google

logging.warn(getversion('GAE DISPATCH'))

bot = WebBot()

class Dispatch_Handler(RequestHandler):

    """ the bots remote command dispatcher. """

    def options(self):
         self.response.headers.add_header('Content-Type', 'application/x-www-form-urlencoded')
         #self.response.headers.add_header("Cache-Control", "private")
         self.response.headers.add_header("Server", getversion())
         self.response.headers.add_header("Public", "*")
         self.response.headers.add_header('Accept', '*')
         self.response.headers.add_header('Access-Control-Allow-Origin', self.request.headers['Origin'])
         self.response.out.write("Allow: *")
         self.response.out.write('Access-Control-Allow-Origin: *') 
         logging.warn("dispatch - options response send to %s - %s" % (self.request.remote_addr, str(self.request.headers)))

    def post(self):

        """ this is where the command get disaptched. """
        starttime = time.time()
        try:
            logging.warn("DISPATCH incoming: %s - %s" % (self.request.get('content'), self.request.remote_addr))
            if not gusers.get_current_user():
                logging.warn("denied access for %s - %s" % (self.request.remote_addr, self.request.get('content')))
                self.response.out.write("acess denied .. plz login")
                self.response.set_status(400)
                return
            event = WebEvent(bot=bot).parse(self.response, self.request)
            event.cbtype = "DISPATCH"
            event.type = "DISPATCH"
            (userhost, user, u, nick) = checkuser(self.response, self.request, event)
            bot.gatekeeper.allow(userhost)
            event.bind(bot)
            bot.doevent(event)
        except NoSuchCommand:
            self.response.out.write("no such command: %s" % event.usercmnd)
        except google.appengine.runtime.DeadlineExceededError, ex:
            self.response.out.write("the command took too long to finish: %s" % str(time.time()-starttime))
        except Exception, ex:
            self.response.out.write("the command had an eror: %s" % exceptionmsg())
            handle_exception()

    get = post

# the application 

application = WSGIApplication([Route('/dispatch/', Dispatch_Handler) ], debug=True)

def main():
    global bot
    global application
    try: application.run()
    except google.appengine.runtime.DeadlineExceededError:
        pass
    except Exception, ex:
        logging.error("dispatch - %s" % str(ex))

if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
#

import os.path

if os.path.isfile("/etc/debian_version") and os.path.isdir("/var/lib/me2"):
    target = "/var/lib/me2"
else:
    target = "me2"

import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

upload = []

def uploadfiles(dir):
    upl = []
    if not os.path.isdir(dir): print "%s does not exist" % dir ; os._exit(1)
    for file in os.listdir(dir):
        if not file or file.startswith('.'):
            continue
        d = dir + os.sep + file
        if not os.path.isdir(d):
            if file.endswith(".pyc"):
                continue
            upl.append(d)

    return upl

def uploadlist(dir):
    upl = []

    for file in os.listdir(dir):
        if not file or file.startswith('.'):
            continue
        d = dir + os.sep + file
        if os.path.isdir(d):   
            upl.extend(uploadlist(d))
        else:
            if file.endswith(".pyc"):
                continue
            upl.append(d)

    return upl

setup(
    name='me2',
    version='0.1a1',
    url='http://botme2.googlecode.com/',
    download_url="http://code.google.com/p/botme2/downloads", 
    author='Bart Thate',
    author_email='bthate@gmail.com',
    description='The bot for you!',
    license='MIT',
    scripts=['bin/me2',
             'bin/me2-init',
             'bin/me2-irc', 
             'bin/me2-fleet', 
             'bin/me2-xmpp', 
             'bin/me2-release',
             'bin/me2-rollback',
             'bin/me2-run',
             'bin/me2-stop',
             'bin/me2-udp',
             'bin/me2-upload',
             'bin/me2-uploadall'],
    packages=['me2',
              'me2.lib', 
              'me2.utils', 
              'me2.lib.console',
              'me2.lib.gae',
              'me2.lib.gae.utils',
              'me2.lib.gae.web',
              'me2.lib.gae.wave',
              'me2.lib.gae.xmpp',
              'me2.lib.socklib',
              'me2.lib.socklib.irc',
              'me2.lib.socklib.xmpp',
              'me2.lib.socklib.utils',
              'me2.lib.socklib.rest',
              'me2.contrib',
              'me2.contrib.simplejson',
              'me2.contrib.tweepy',
              'me2.plugs',
              'me2.plugs.core',
              'me2.plugs.wave',
              'me2.plugs.common',
              'me2.plugs.socket', 
              'me2.plugs.gae',
              'me2.plugs.myplugs'],
    long_description = """ BOTME2 is a remote event-driven framework for building bots that talk JSON to each other over XMPP. IRC/Console/XMPP (shell) Wave/Web/XMPP (GAE) implementations provided. """,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: Other OS',
        'Programming Language :: Python',
        'Topic :: Communications :: Chat',
        'Topic :: Software Development :: Libraries :: Python Modules'],
    data_files=[(target + os.sep + 'data', uploadlist(target + os.sep + 'data')),
                (target + os.sep + 'data' + os.sep + 'examples', uploadfiles('me2' + os.sep + 'data' + os.sep + 'examples')),
                (target + os.sep + 'upload', uploadfiles('me2' + os.sep + 'upload')),
                (target + os.sep + 'upload' + os.sep + 'webapp2', uploadlist('me2' + os.sep + 'upload' + os.sep + 'webapp2')),
                (target + os.sep + 'upload' + os.sep + 'assets', uploadlist('me2' + os.sep + 'upload' + os.sep + 'assets')),
                (target + os.sep + 'upload' + os.sep + 'templates', uploadlist('me2' + os.sep + 'upload' + os.sep +'templates')),
                (target + os.sep + 'upload' + os.sep + 'waveapi', uploadlist('me2' + os.sep + 'upload' + os.sep + 'waveapi')),
                (target + os.sep + 'upload' + os.sep + 'waveapi' + os.sep + 'oauth', uploadlist('me2' + os.sep + 'upload' + os.sep + 'waveapi' + os.sep + 'oauth')),
                (target + os.sep + 'upload' + os.sep + 'waveapi' + os.sep + 'simplejson', uploadlist('me2' + os.sep + 'upload' + os.sep + 'waveapi' + os.sep + 'simplejson')),
                (target + os.sep + 'upload' + os.sep + 'gadgets', uploadlist('me2' + os.sep + 'upload' + os.sep + 'gadgets'))],
)



# me2/socklib/xmpp/bot.py
#
#

""" jabber bot definition """

## me2 imports

from me2.lib.users import users
from me2.utils.exception import handle_exception
from me2.utils.trace import whichmodule
from me2.utils.locking import lockdec
from me2.utils.pdod import Pdod
from me2.utils.dol import Dol
from me2.lib.less import Less
from me2.lib.callbacks import callbacks, remote_callbacks
from me2.lib.threads import start_new_thread
from me2.lib.botbase import BotBase
from me2.lib.exit import globalshutdown
from me2.lib.channelbase import ChannelBase
from me2.lib.fleet import getfleet

## me2.socket imports

from me2.lib.socklib.utils.generic import waitforqueue, jabberstrip, getrandomnick
from me2.utils.generic import toenc, fromenc

## xmpp imports

from me2.contrib.xmlstream import XMLescape, XMLunescape
from presence import Presence
from message import Message
from iq import Iq
from core import XMLStream
from wait import XMPPWait, XMPPErrorWait
from jid import JID, InvalidJID
from errors import xmpperrors

## basic imports

import time
import Queue
import os
import threading
import thread
import types
import xml
import re
import hashlib
import logging
import cgi

## locks

outlock = thread.allocate_lock()
inlock = thread.allocate_lock()
connectlock = thread.allocate_lock()
outlocked = lockdec(outlock)
inlocked = lockdec(inlock)
connectlocked = lockdec(connectlock)

## SXMPPBot class

class SXMPPBot(XMLStream, BotBase):

    """
        xmpp bot class.

    """

    def __init__(self, cfg=None, usersin=None, plugs=None, jid=None, *args, **kwargs):
        BotBase.__init__(self, cfg, usersin, plugs, jid, *args, **kwargs)
        self.port = 5222
        if not self.host:
            self.host = self.cfg.host
            if not self.host: raise Exception("%s - host not set - %s" % (self.name, str(self.cfg)))
        self.username = self.user.split('@')[0]
        XMLStream.__init__(self, self.host, self.port, self.name)   
        self.type = 'sxmpp'
        self.sock = None
        self.me = self.cfg.user
        self.jid = self.me
        self.lastin = None
        self.test = 0
        self.password = ""
        self.connecttime = 0
        self.connection = None
        self.privwait = XMPPWait()
        self.errorwait = XMPPErrorWait()
        self.jabber = True
        self.jids = {}
        self.topics = {}
        self.timejoined = {}
        self.channels409 = []
        if self.state and not self.state.data.ratelimit: self.state.data.ratelimit = 0.05
        if self.port == 0: self.port = 5222

    def _resumedata(self):
        """ return data needed for resuming. """
        return {self.name: {
            'type': self.type,
            'nick': self.nick,
            'server': self.server,
            'port': self.port,
            'password': self.password,
            'ipv6': self.ipv6,
            'user': self.user
            }}

    def _keepalive(self):
        """ keepalive method .. send empty string to self every 3 minutes. """
        nrsec = 0
        while not self.stopped:
            time.sleep(1)
            nrsec += 1
            if nrsec < 180: continue
            else: nrsec = 0
            self.sendpresence()

    def sendpresence(self):
        """ send presence based on status and status text set by user. """
        if self.state:
            if self.state.has_key('status') and self.state['status']: status = self.state['status']
            else: status = ""
            if self.state.has_key('show') and self.state['show']: show = self.state['show']
            else: show = ""
        else:
            status = ""
            show = ""
        logging.debug('%s - keepalive - %s - %s' % (self.name, show, status))
        if show and status: p = Presence({'to': self.me, 'show': show, 'status': status})
        elif show: p = Presence({'to': self.me, 'show': show })
        elif status: p = Presence({'to': self.me, 'status': status})
        else: p = Presence({'to': self.me })
        self.send(p)

    def _keepchannelsalive(self):
        """ channels keep alive method. """
        nrsec = 0
        p = Presence({'to': self.me, 'txt': '' })
        while not self.stopped:
            time.sleep(1)
            nrsec += 1
            if nrsec < 600: continue
            else: nrsec = 0
            for chan in self.state['joinedchannels']:
                if chan not in self.channels409:
                    p = Presence({'to': chan})
                    self.send(p)

    def connect(self, reconnect=True):
        """ connect the xmpp server. """
        try:
            if not XMLStream.connect(self):
                logging.error('%s - connect to %s:%s failed' % (self.name, self.host, self.port))
                return
            else: logging.warn('%s - connected' % self.name)
            self.logon(self.cfg.user, self.cfg.password)
            start_new_thread(self._keepalive, ())
            self.requestroster()
            self._raw("<presence/>")
            self.connectok.set()
            self.sock.settimeout(None)
            return True
        except Exception, ex:
            handle_exception()
            if reconnect:
                return self.reconnect()

    def logon(self, user, password):
        """ logon on the xmpp server. """
        iq = self.initstream()
        if not self.auth(user, password, iq.id):
            logging.warn("%s - sleeping 20 seconds before register" % self.name)
            time.sleep(20)
            if self.register(user, password):
                time.sleep(5)
                self.auth(user, password)
            else:
                time.sleep(10)
                self.exit()
                return
        XMLStream.logon(self)
 
    def initstream(self):
        """ send initial string sequence to the xmpp server. """
        logging.debug('%s - starting initial stream sequence' % self.name)
        self._raw("""<stream:stream to='%s' xmlns='jabber:client' xmlns:stream='http://etherx.jabber.org/streams'>""" % (self.user.split('@')[1], )) 
        result = self.connection.read()
        iq = self.loop_one(result)
        logging.debug("%s - initstream - %s" % (self.name, result))
        return iq

    def register(self, jid, password):
        """ register the jid to the server. """
        try: resource = jid.split("/")[1]
        except IndexError: resource = "me2"
        logging.warn('%s - registering %s' % (self.name, jid))
        self._raw("""<iq type='get'><query xmlns='jabber:iq:register'/></iq>""")
        result = self.connection.read()
        iq = self.loop_one(result)
        if not iq:
            logging.error("%s - unable to register" % self.name)
            return
        logging.debug('%s - register: %s' % (self.name, str(iq)))
        self._raw("""<iq type='set'><query xmlns='jabber:iq:register'><username>%s</username><resource>%s</resource><password>%s</password></query></iq>""" % (jid.split('@')[0], resource, password))
        result = self.connection.read()
        logging.debug('%s - register - %s' % (self.name, result))
        if not result: return False
        iq = self.loop_one(result)
        if not iq:
            logging.error("%s - can't decode data - %s" % (self.name, result))
            return False
        logging.debug('sxmpp - register - %s' % result)
        if iq.error:
            logging.warn('%s - register FAILED - %s' % (self.name, iq.error))
            if not iq.error.code: logging.error("%s - can't determine error code" % self.name) ; return False
            if iq.error.code == "405": logging.error("%s - this server doesn't allow registration by the bot, you need to create an account for it yourself" % self.name)
            elif iq.error.code == "500": logging.error("%s - %s" % (self.name, iq.error.text.data))
            else: logging.error("%s - %s" % (self.name, xmpperrors[iq.error.code]))
            self.error = iq.error
            return False
        logging.warn('%s - register ok' % self.name)
        return True

    def auth(self, jid, password, digest=""):
        """ auth against the xmpp server. """
        logging.warn('%s - authing %s' % (self.name, jid))
        name = jid.split('@')[0]
        rsrc = self.cfg['resource'] or self.cfg['resource'] or 'me2';
        self._raw("""<iq type='get'><query xmlns='jabber:iq:auth'><username>%s</username></query></iq>""" % name)
        result = self.connection.read()
        iq = self.loop_one(result)
        logging.debug('%s - auth - %s' % (self.name, result))
        if ('digest' in result) and digest:
            s = hashlib.new('SHA1')
            s.update(digest)
            s.update(password)
            d = s.hexdigest()
            self._raw("""<iq type='set'><query xmlns='jabber:iq:auth'><username>%s</username><digest>%s</digest><resource>%s</resource></query></iq>""" % (name, d, rsrc))
        else: self._raw("""<iq type='set'><query xmlns='jabber:iq:auth'><username>%s</username><resource>%s</resource><password>%s</password></query></iq>""" % (name, rsrc, password))
        result = self.connection.read()
        iq = self.loop_one(result)
        if not iq:
            logging.error('%s - auth failed - %s' % (self.name, result))
            return False        
        logging.debug('%s - auth - %s' % (self.name, result))
        if iq.error:
            logging.warn('%s - auth failed - %s' % (self.name, iq.error))
            if iq.error.code == "401":
                logging.warn("%s - wrong user or password" % self.name)
            else:
                logging.warn("%s - %s" % (self.name, result))
            self.error = iq.error
            return False
        logging.warn('%s - auth ok' % self.name)
        return True

    def requestroster(self):
        """ request roster from xmpp server. """
        self._raw("<iq type='get'><query xmlns='jabber:iq:roster'/></iq>")

    def disconnectHandler(self, ex):
        """ disconnect handler. """
        self.reconnect()

    def outnocb(self, printto, txt, how=None, event=None, html=False, *args, **kwargs):
        """ output txt to bot. """
        if printto and printto in self.state['joinedchannels']: outtype = 'groupchat'
        else: outtype = "chat"
        target = printto
        if not html: 
            txt = self.normalize(txt)
        repl = Message({'from': self.me, 'to': target, 'type': outtype, 'txt': txt})
        if html:
            repl.html = txt
        if not repl.type: repl.type = 'normal'
        logging.debug("%s - sxmpp - out - %s - %s" % (self.name, printto, unicode(txt)))
        self.send(repl)

    def broadcast(self, txt):
        """ broadcast txt to all joined channels. """
        for i in self.state['joinedchannels']:
            self.say(i, txt)

    def handle_iq(self, data):
        """ iq handler .. overload this when needed. """
        pass

    def handle_message(self, data):
        """ message handler. """
        m = Message(data)
        m.parse(self)
        if data.type == 'groupchat' and data.subject:
            logging.debug("%s - checking topic" % self.name)
            self.topiccheck(m)
            nm = Message(m)
            callbacks.check(self, nm)
            return
        if data.get('x').xmlns == 'jabber:x:delay':
            logging.warn("%s - ignoring delayed message" % self.name)
            return
        self.privwait.check(m)
        if m.isresponse:
            logging.debug("%s - message is a response" % self.name)
            return
        jid = None
        m.origjid = m.jid
        for node in m.subelements:
            try: m.jid = node.x.item.jid 
            except (AttributeError, TypeError): continue
        if self.me in m.fromm:
            logging.debug("%s - message to self .. ignoring" % self.name)
            return 0
        try:
            if m.type == 'error':
                if m.code:
                    logging.error('%s - error - %s' % (self.name, str(m)))
                self.errorwait.check(m)
                self.errorHandler(m)
        except Exception, ex:
            handle_exception()
        self.put(m)

    def errorHandler(self, event):
        """ error handler .. calls the errorhandler set in the event. """
        try:
            logging.error("%s - error occured in %s" % (self.name, event.dump()))
            event.errorHandler()
        except AttributeError: logging.error('%s - unhandled error - %s' % (self.name, event.dump()))

    def handle_presence(self, data):
        """ presence handler. """
        p = Presence(data)
        p.parse()
        frm = p.fromm
        nickk = ""
        nick = p.nick
        if self.me in p.userhost: return 0
        if nick: 
            self.userhosts[nick] = str(frm)
            nickk = nick
        jid = None
        for node in p.subelements:
            try:
                jid = node.x.item.jid 
            except (AttributeError, TypeError):
                continue
        if nickk and jid:
            channel = p.channel
            if not self.jids.has_key(channel):
                self.jids[channel] = {}
            self.jids[channel][nickk] = jid
            self.userhosts[nickk] = str(jid)
            logging.debug('%s - setting jid of %s (%s) to %s' % (self.name, nickk, channel, jid))
        if p.type == 'subscribe':
            pres = Presence({'to': p.fromm, 'type': 'subscribed'})
            self.send(pres)
            pres = Presence({'to': p.fromm, 'type': 'subscribe'})
            self.send(pres)
        nick = p.resource
        if p.type != 'unavailable':
            p.joined = True
            p.type = 'available'
        elif self.me in p.userhost:
            try:
                del self.jids[p.channel]
                logging.debug('%s - removed %s channel jids' % (self.name, p.channel))
            except KeyError:
                pass
        else:
            try:
                del self.jids[p.channel][p.nick]
                logging.debug('%s - removed %s jid' % (self.name, p.nick))
            except KeyError:
                pass
        if p.type == 'error':
            for node in p.subelements:
                try:
                    err = node.error.code
                except (AttributeError, TypeError):
                    err = 'no error set'
                try:
                    txt = node.text.data
                except (AttributeError, TypeError):
                    txt = ""
            if err:
                logging.error('%s - error - %s - %s'  % (self.name, err, txt))
            self.errorwait.check(p)
            try:
                method = getattr(self,'handle_' + err)
                try:
                    method(p)
                except:
                    handle_exception()
            except AttributeError:
                pass
        self.doevent(p)

    def invite(self, jid):
        pres = Presence({'to': jid, 'type': 'subscribe'})
        self.send(pres)
        time.sleep(2)
        pres = Presence({'to': jid})
        self.send(pres)

    def send(self, what):
        """ send stanza to the server. """
        if not what:
            logging.debug("%s - can't send empty message" % self.name)
            return
        try:
            to = what['to']
        except (KeyError, TypeError):
            logging.error("%s - can't determine where to send %s to" % (self.name, what))
            return
        try:
            jid = JID(to)
        except (InvalidJID, AttributeError):
            logging.error("%s - invalid jid - %s - %s" % (self.name, str(to), str(what)))
            return
        try: del what['from']
        except KeyError: pass
        try:
            xml = what.tojabber()
            if not xml:
                raise Exception("can't convert %s to xml .. bot.send()" % what) 
        except (AttributeError, TypeError):
            handle_exception()
            return
        if not self.checkifvalid(xml): logging.error("%s - NOT PROPER XML - %s" % (self.name, xml))
        else: self._raw(xml)
           
    def action(self, printto, txt, fromm=None, groupchat=True):
        """ send an action. """
        txt = "/me " + txt
        if self.google:
            fromm = self.me
        if printto in self.state['joinedchannels'] and groupchat:
            message = Message({'to': printto, 'txt': txt, 'type': 'groupchat'})
        else: message = Message({'to': printto, 'txt': txt})
        if fromm: message.fromm = fromm
        self.send(message)
        
    def userwait(self, msg, txt):
        """ wait for user response. """
        msg.reply(txt)
        queue = Queue.Queue()
        self.privwait.register(msg, queue)
        result = queue.get()
        if result:
            return result.txt

    def save(self):
        """ save bot's state. """
        if self.state:
            self.state.save()

    def quit(self):
        """ send unavailable presence. """
        if self.error: return
        presence = Presence({'type': 'unavailable' ,'to': self.jid})
        if self.state:
            for i in self.state.data.joinedchannels:
                presence.to = i
                self.send(presence)
        presence = Presence({'type': 'unavailable', 'to': self.jid})
        presence['from'] = self.me
        self.send(presence)

    def shutdown(self):
        self.outqueue.put_nowait(None)

    def join(self, channel, password=None, nick=None):
        """ join conference. """
        if channel.startswith("#"): return
        try:
            if not nick: nick = channel.split('/')[1]
        except IndexError: nick = self.nick
        channel = channel.split('/')[0]
        q = Queue.Queue()
        self.errorwait.register("409", q, 3)
        self.errorwait.register("401", q, 3)
        self.errorwait.register("400", q, 3)
        presence = Presence({'to': channel + '/' + nick})
        if password:
             presence.x.password = password             
        self.send(presence)
        errorobj = waitforqueue(q, 3)
        if errorobj:
            err = errorobj[0].error
            logging.error('%s - error joining %s - %s' % (self.name, channel, err))
            if err >= '400':
                if channel not in self.channels409: self.channels409.append(channel)
            return err
        self.timejoined[channel] = time.time()
        chan = ChannelBase(channel, self.botname)
        chan.data['nick'] = nick
        if password:
            chan.data['key'] = password
        if not chan.data.has_key('cc'):
            chan.data['cc'] = self.cfg['defaultcc'] or '!'
        if channel not in self.state['joinedchannels']:
            self.state['joinedchannels'].append(channel)
            self.state.save()
        if channel in self.channels409:
            self.channels409.remove(channel)
        return 1

    def part(self, channel):
        """ leave conference. """
        if channel.startswith("#"): return
        presence = Presence({'to': channel})
        presence.type = 'unavailable'
        self.send(presence)
        if channel in self.state['joinedchannels']: self.state['joinedchannels'].remove(channel)
        self.state.save()
        return 1

    def outputnolog(self, printto, what, how, who=None, fromm=None):
        """ do output but don't log it. """
        if fromm: return
        self.saynocb(printto, what)

    def topiccheck(self, msg):
        """ check if topic is set. """
        if msg.groupchat:
            try:
                topic = msg.subject
                if not topic: return None
                self.topics[msg.channel] = (topic, msg.userhost, time.time())
                logging.debug('%s - topic of %s set to %s' % (self.name, msg.channel, topic))
            except AttributeError: return None

    def settopic(self, channel, txt):
        """ set topic. """
        pres = Message({'to': channel, 'subject': txt})
        pres.type = 'groupchat'
        self.send(pres)

    def gettopic(self, channel):
        """ get topic. """
        try:
            topic = self.topics[channel]
            return topic
        except KeyError: return None

    def domsg(self, msg):
        """ dispatch an msg on the bot. """
        self.doevent(msg)

    def normalize(self, what):
        #what = cgi.escape(what)
        what = what.replace("\002", "")
        what = what.replace("<b>", "")
        what = what.replace("</b>", "")
        what = what.replace("&lt;b&gt;", "")
        what = what.replace("&lt;/b&gt;", "")
        what = what.replace("<i>", "")
        what = what.replace("</i>", "")
        what = what.replace("&lt;i&gt;", "")
        what = what.replace("&lt;/i&gt;", "")
        return what

    def doreconnect(self):
        """ reconnect to the server. """
        botjid = self.jid
        newbot = getfleet().makebot('sxmpp', self.name, cfg=self.cfg)
        newbot.reconnectcount = self.reconnectcount
        self.exit()
        if newbot.start():
            self.jid += '.old'
            #newbot.joinchannels()
            if fleet.replace(botjid, newbot): return True
        return False




# me2/plugs/welcome.py
#
#

""" send welcome message. """

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.examples import examples

## welcome command

def handle_welcome(bot, event):
    event.reply("Welcome to BOTME2 .. The JSON everywhere bot ;] for wave/web/xmpp/IRC/console - you can give this bot commands. try !help.")

cmnds.add('welcome', handle_welcome, ['USER', 'GUEST'])
examples.add('welcome', 'send welcome msg', 'welcome')



# plugs/gcalc.py
# encoding: utf-8
#
#

""" use google to calculate e.g. !gcalc 1 + 1 """

## me2 imports

from me2.lib.commands import cmnds
from me2.lib.examples import examples
from me2.utils.url import useragent

from me2.contrib.simplejson import loads

## basic imports

import urllib2

## gcalc command

def handle_gcalc(bot, ievent):
    """ use google calc . """
    if len(ievent.args) > 0: expr = " ".join(ievent.args).replace("+", "%2B").replace(" ", "+")
    else: ievent.missing('Missing an expression') ; return
    req = urllib2.Request("http://www.google.com/ig/calculator?hl=en&q=%s" % expr, None,  {'User-agent': useragent()})
    data = urllib2.urlopen(req).read()
    try:
        rhs = data.split("rhs")[1].split("\"")[1]
        lhs = data.split("lhs")[1].split("\"")[1]
        ievent.reply("%s = %s" % (lhs,rhs.replace('\\x26#215;', '*').replace('\\x3csup\\x3e', '**').replace('\\x3c/sup\\x3e', '')))
    except Exception, ex:
        ievent.reply(str(ex))    

cmnds.add('gcalc', handle_gcalc, ['USER', 'GUEST'])
examples.add('gcalc', 'calculate an expression using the google calculator', 'gcalc 1 + 1')
